namespace com.progsbase.no.inductive.libraries.FormulaTranslation_v0_1_9{

using static System.Math;

public class ASTNode{
	public char [] value;
	public bool leaf;
	public ASTNode l;
	public ASTNode r;
}
public class BooleanArrayReference{
	public bool [] booleanArray;
}
public class BooleanReference{
	public bool booleanValue;
}
public class CharacterReference{
	public char characterValue;
}
public class NumberArrayReference{
	public double [] numberArray;
}
public class NumberReference{
	public double numberValue;
}
public class StringArrayReference{
	public StringReference [] stringArray;
}
public class StringReference{
	public char [] stringx;
}
public class DynamicArrayCharacters{
	public char [] array;
	public double length;
}
public class LinkedListNodeStrings{
	public bool end;
	public char [] value;
	public LinkedListNodeStrings next;
}
public class LinkedListStrings{
	public LinkedListNodeStrings first;
	public LinkedListNodeStrings last;
}
public class LinkedListNodeNumbers{
	public LinkedListNodeNumbers next;
	public bool end;
	public double value;
}
public class LinkedListNumbers{
	public LinkedListNodeNumbers first;
	public LinkedListNodeNumbers last;
}
public class LinkedListCharacters{
	public LinkedListNodeCharacters first;
	public LinkedListNodeCharacters last;
}
public class LinkedListNodeCharacters{
	public bool end;
	public char value;
	public LinkedListNodeCharacters next;
}
public class DynamicArrayNumbers{
	public double [] array;
	public double length;
}
public class FormulaTranslation{
	public static bool TokenizeBooleanFormula(char [] f, StringArrayReference tokens, StringReference errorMessage){
		bool success;
		double i;
		LinkedListStrings ll;
		LinkedListCharacters llc;
		char [] str;
		bool found;
		char c1, c2, c3;

		success = true;
		ll = CreateLinkedListString();
		llc = CreateLinkedListCharacter();

		for(i = 0d; i < f.Length && success; ){
			found = false;

			c1 = f[(int)(i)];

			if(i + 2d < f.Length){
				c2 = f[(int)(i + 1d)];
				c3 = f[(int)(i + 2d)];

				if(c1 == '<' && c2 == '-' && c3 == '>'){
					LinkedListAddString(ll, "<->".ToCharArray());
					i = i + 3d;
					found = true;
				}
			}

			if(!found){
				if(i + 1d < f.Length){
					c2 = f[(int)(i + 1d)];

					if(c1 == '<' && c2 == '='){
						LinkedListAddString(ll, "<=".ToCharArray());
						i = i + 2d;
						found = true;
					}else if(c1 == '>' && c2 == '='){
						LinkedListAddString(ll, ">=".ToCharArray());
						i = i + 2d;
						found = true;
					}else if(c1 == '!' && c2 == '='){
						LinkedListAddString(ll, "!=".ToCharArray());
						i = i + 2d;
						found = true;
					}else if(c1 == '-' && c2 == '>'){
						LinkedListAddString(ll, "->".ToCharArray());
						i = i + 2d;
						found = true;
					}
				}
			}

			if(!found){
				if(c1 == '('){
					LinkedListAddString(ll, "(".ToCharArray());
					i = i + 1d;
					found = true;
				}else if(c1 == ')'){
					LinkedListAddString(ll, ")".ToCharArray());
					i = i + 1d;
					found = true;
				}else if(c1 == '!'){
					LinkedListAddString(ll, "!".ToCharArray());
					i = i + 1d;
					found = true;
				}else if(c1 == '&'){
					LinkedListAddString(ll, "&".ToCharArray());
					i = i + 1d;
					found = true;
				}else if(c1 == '^'){
					LinkedListAddString(ll, "^".ToCharArray());
					i = i + 1d;
					found = true;
				}else if(c1 == '|'){
					LinkedListAddString(ll, "|".ToCharArray());
					i = i + 1d;
					found = true;
				}else if(c1 == '='){
					LinkedListAddString(ll, "=".ToCharArray());
					i = i + 1d;
					found = true;
				}else if(c1 == '<'){
					LinkedListAddString(ll, "<".ToCharArray());
					i = i + 1d;
					found = true;
				}else if(c1 == '>'){
					LinkedListAddString(ll, ">".ToCharArray());
					i = i + 1d;
					found = true;
				}else if(c1 == ' ' || c1 == '\t' || c1 == '\n'){
					i = i + 1d;
					found = true;
				}else if(charIsLetter(c1)){
					for(; i < f.Length && (charIsLetter(f[(int)(i)]) || charIsNumber(f[(int)(i)]) || f[(int)(i)] == '_'); i = i + 1d){
						LinkedListAddCharacter(llc, f[(int)(i)]);
					}
					str = LinkedListCharactersToArray(llc);
					LinkedListAddString(ll, str);
					FreeLinkedListCharacter(llc);
					llc = CreateLinkedListCharacter();

					found = true;
				}else if(charIsNumber(c1)){
					for(; i < f.Length && (charIsNumber(f[(int)(i)]) || f[(int)(i)] == '.'); i = i + 1d){
						LinkedListAddCharacter(llc, f[(int)(i)]);
					}
					str = LinkedListCharactersToArray(llc);
					LinkedListAddString(ll, str);
					FreeLinkedListCharacter(llc);
					llc = CreateLinkedListCharacter();

					found = true;
				}
			}

			if(!found){
				success = false;
				errorMessage.stringx = "Unknown token in formula.".ToCharArray();
			}
		}

		FreeLinkedListCharacter(llc);
		if(success){
			LinkedListAddString(ll, "<end>".ToCharArray());
			tokens.stringArray = LinkedListStringsToArray(ll);
		}
		FreeLinkedListString(ll);

		return success;
	}


	public static bool ParseBooleanTokens(StringArrayReference tokens, NumberReference cur, ASTNode ast, StringReference errorMessage){
		bool success;

		success = ParseIff(tokens, cur, ast, errorMessage);

		if(success){
			success = TokenIs(tokens, cur, "<end>".ToCharArray());

			if(!success){
				errorMessage.stringx = "Expected the end of the formula.".ToCharArray();
			}
		}

		return success;
	}


	public static bool ParseIff(StringArrayReference tokens, NumberReference cur, ASTNode ast, StringReference errorMessage){
		bool success;
		ASTNode t, t1;
		char [] op;

		t = new ASTNode();
		success = ParseImplies(tokens, cur, t, errorMessage);

		for(; success && TokenIs(tokens, cur, "<->".ToCharArray()); ){
			op = Index(tokens, cur);
			AddToNumberReference(cur, 1d);
			t1 = new ASTNode();
			success = ParseImplies(tokens, cur, t1, errorMessage);
			if(success){
				t = CreateASTNode(t, t1, op);
			}
		}

		if(success){
			AssignASTNode(ast, t);
		}

		return success;
	}


	public static bool ParseImplies(StringArrayReference tokens, NumberReference cur, ASTNode ast, StringReference errorMessage){
		bool success;
		ASTNode t, t1;
		char [] op;

		t = new ASTNode();
		success = ParseLogicalOr(tokens, cur, t, errorMessage);

		for(; success && TokenIs(tokens, cur, "->".ToCharArray()); ){
			op = Index(tokens, cur);
			AddToNumberReference(cur, 1d);
			t1 = new ASTNode();
			success = ParseLogicalOr(tokens, cur, t1, errorMessage);
			if(success){
				t = CreateASTNode(t, t1, op);
			}
		}

		if(success){
			AssignASTNode(ast, t);
		}

		return success;
	}


	public static bool ParseLogicalOr(StringArrayReference tokens, NumberReference cur, ASTNode ast, StringReference errorMessage){
		bool success;
		ASTNode t, t1;
		char [] op;

		t = new ASTNode();
		success = ParseLogicalAnd(tokens, cur, t, errorMessage);

		for(; success && TokenIs(tokens, cur, "|".ToCharArray()); ){
			op = Index(tokens, cur);
			AddToNumberReference(cur, 1d);
			t1 = new ASTNode();
			success = ParseLogicalAnd(tokens, cur, t1, errorMessage);
			if(success){
				t = CreateASTNode(t, t1, op);
			}
		}

		if(success){
			AssignASTNode(ast, t);
		}

		return success;
	}


	public static bool ParseLogicalAnd(StringArrayReference tokens, NumberReference cur, ASTNode ast, StringReference errorMessage){
		bool success;
		ASTNode t, t1;
		char [] op;

		t = new ASTNode();
		success = ParseLogicalXor(tokens, cur, t, errorMessage);

		for(; success && TokenIs(tokens, cur, "&".ToCharArray()); ){
			op = Index(tokens, cur);
			AddToNumberReference(cur, 1d);
			t1 = new ASTNode();
			success = ParseLogicalXor(tokens, cur, t1, errorMessage);
			if(success){
				t = CreateASTNode(t, t1, op);
			}
		}

		if(success){
			AssignASTNode(ast, t);
		}

		return success;
	}


	public static bool ParseLogicalXor(StringArrayReference tokens, NumberReference cur, ASTNode ast, StringReference errorMessage){
		bool success;
		ASTNode t, t1;
		char [] op;

		t = new ASTNode();
		success = ParseLogicalComparison(tokens, cur, t, errorMessage);

		for(; success && TokenIs(tokens, cur, "^".ToCharArray()); ){
			op = Index(tokens, cur);
			AddToNumberReference(cur, 1d);
			t1 = new ASTNode();
			success = ParseLogicalComparison(tokens, cur, t1, errorMessage);
			if(success){
				t = CreateASTNode(t, t1, op);
			}
		}

		if(success){
			AssignASTNode(ast, t);
		}

		return success;
	}


	public static bool ParseLogicalComparison(StringArrayReference tokens, NumberReference cur, ASTNode ast, StringReference errorMessage){
		bool success;
		ASTNode t, t1;
		char [] op;

		t = new ASTNode();
		success = ParseLogicalLiteralVariableNotParenthesis(tokens, cur, t, errorMessage);

		for(; success && (TokenIs(tokens, cur, "=".ToCharArray()) || TokenIs(tokens, cur, "!=".ToCharArray()) || TokenIs(tokens, cur, "<".ToCharArray()) || TokenIs(tokens, cur, "<=".ToCharArray()) || TokenIs(tokens, cur, ">".ToCharArray()) || TokenIs(tokens, cur, ">=".ToCharArray())); ){
			op = Index(tokens, cur);
			AddToNumberReference(cur, 1d);
			t1 = new ASTNode();
			success = ParseLogicalLiteralVariableNotParenthesis(tokens, cur, t1, errorMessage);
			if(success){
				t = CreateASTNode(t, t1, op);
			}
		}

		if(success){
			AssignASTNode(ast, t);
		}

		return success;
	}


	public static bool ParseLogicalLiteralVariableNotParenthesis(StringArrayReference tokens, NumberReference cur, ASTNode ast, StringReference errorMessage){
		bool success;
		ASTNode t;
		char [] token;

		success = true;

		token = Index(tokens, cur);
		if(charIsLetter(token[0]) || charIsNumber(token[0])){
			ast.leaf = true;
			ast.value = token;
			AddToNumberReference(cur, 1d);
		}else if(TokenIs(tokens, cur, "(".ToCharArray())){
			AddToNumberReference(cur, 1d);

			t = new ASTNode();
			success = ParseIff(tokens, cur, t, errorMessage);

			if(success){
				ast.value = "()".ToCharArray();
				ast.l = t;
				ast.leaf = false;

				success = TokenIs(tokens, cur, ")".ToCharArray());
				if(success){
					AddToNumberReference(cur, 1d);
				}else{
					errorMessage.stringx = "Epected \')\'.".ToCharArray();
				}
			}
		}else if(TokenIs(tokens, cur, "!".ToCharArray())){
			AddToNumberReference(cur, 1d);

			t = new ASTNode();
			success = ParseLogicalLiteralVariableNotParenthesis(tokens, cur, t, errorMessage);

			if(success){
				ast.value = "!".ToCharArray();
				ast.l = t;
				ast.leaf = false;
			}
		}else{
			success = false;
			errorMessage.stringx = "Unexpected token.".ToCharArray();
		}

		return success;
	}


	public static bool TokenizeArithmeticFormula(char [] f, StringArrayReference tokens, StringReference errorMessage){
		bool success;
		double i;
		LinkedListStrings ll;
		LinkedListCharacters llc;
		char [] str;

		success = true;
		ll = CreateLinkedListString();
		llc = CreateLinkedListCharacter();

		for(i = 0d; i < f.Length && success; ){
			if(f[(int)(i)] == '('){
				LinkedListAddString(ll, "(".ToCharArray());
				i = i + 1d;
			}else if(f[(int)(i)] == ')'){
				LinkedListAddString(ll, ")".ToCharArray());
				i = i + 1d;
			}else if(f[(int)(i)] == '+'){
				LinkedListAddString(ll, "+".ToCharArray());
				i = i + 1d;
			}else if(f[(int)(i)] == '-'){
				LinkedListAddString(ll, "-".ToCharArray());
				i = i + 1d;
			}else if(f[(int)(i)] == '*'){
				LinkedListAddString(ll, "*".ToCharArray());
				i = i + 1d;
			}else if(f[(int)(i)] == '/'){
				LinkedListAddString(ll, "/".ToCharArray());
				i = i + 1d;
			}else if(f[(int)(i)] == '^'){
				LinkedListAddString(ll, "^".ToCharArray());
				i = i + 1d;
			}else if(f[(int)(i)] == '%'){
				LinkedListAddString(ll, "%".ToCharArray());
				i = i + 1d;
			}else if(f[(int)(i)] == ' ' || f[(int)(i)] == '\t' || f[(int)(i)] == '\n'){
				i = i + 1d;
			}else if(charIsLetter(f[(int)(i)])){
				for(; i < f.Length && (charIsLetter(f[(int)(i)]) || charIsNumber(f[(int)(i)]) || f[(int)(i)] == '_'); i = i + 1d){
					LinkedListAddCharacter(llc, f[(int)(i)]);
				}
				str = LinkedListCharactersToArray(llc);
				LinkedListAddString(ll, str);
				FreeLinkedListCharacter(llc);
				llc = CreateLinkedListCharacter();
			}else if(charIsNumber(f[(int)(i)])){
				for(; i < f.Length && (charIsNumber(f[(int)(i)]) || f[(int)(i)] == '.'); i = i + 1d){
					LinkedListAddCharacter(llc, f[(int)(i)]);
				}
				str = LinkedListCharactersToArray(llc);
				LinkedListAddString(ll, str);
				FreeLinkedListCharacter(llc);
				llc = CreateLinkedListCharacter();
			}else{
				success = false;
				errorMessage.stringx = "Unknown token in formula.".ToCharArray();
			}
		}

		FreeLinkedListCharacter(llc);
		if(success){
			LinkedListAddString(ll, "<end>".ToCharArray());
			tokens.stringArray = LinkedListStringsToArray(ll);
		}
		FreeLinkedListString(ll);

		return success;
	}


	public static bool ParseArithmeticTokens(StringArrayReference tokens, NumberReference cur, ASTNode ast, StringReference errorMessage){
		bool success;

		success = ParseAdditionOrSubtraction(tokens, cur, ast, errorMessage);

		if(success){
			success = TokenIs(tokens, cur, "<end>".ToCharArray());

			if(!success){
				errorMessage.stringx = "Expected the end of the formula.".ToCharArray();
			}
		}

		return success;
	}


	public static bool ParseAdditionOrSubtraction(StringArrayReference tokens, NumberReference cur, ASTNode ast, StringReference errorMessage){
		bool success;
		ASTNode t, t1;
		char [] op;

		t = new ASTNode();
		success = ParseMultiplicationOrDivision(tokens, cur, t, errorMessage);

		for(; success && (TokenIs(tokens, cur, "+".ToCharArray()) || TokenIs(tokens, cur, "-".ToCharArray())); ){
			op = Index(tokens, cur);
			AddToNumberReference(cur, 1d);
			t1 = new ASTNode();
			success = ParseMultiplicationOrDivision(tokens, cur, t1, errorMessage);
			if(success){
				t = CreateASTNode(t, t1, op);
			}
		}

		if(success){
			AssignASTNode(ast, t);
		}

		return success;
	}


	public static bool ParseMultiplicationOrDivision(StringArrayReference tokens, NumberReference cur, ASTNode ast, StringReference errorMessage){
		bool success;
		ASTNode t, t1;
		char [] op;

		t = new ASTNode();
		success = ParseExponentiation(tokens, cur, t, errorMessage);

		for(; success && (TokenIs(tokens, cur, "*".ToCharArray()) || TokenIs(tokens, cur, "/".ToCharArray()) || TokenIs(tokens, cur, "mod".ToCharArray())); ){
			op = Index(tokens, cur);
			AddToNumberReference(cur, 1d);
			t1 = new ASTNode();
			success = ParseExponentiation(tokens, cur, t1, errorMessage);
			if(success){
				t = CreateASTNode(t, t1, op);
			}
		}

		if(success){
			AssignASTNode(ast, t);
		}

		return success;
	}


	public static bool ParseExponentiation(StringArrayReference tokens, NumberReference cur, ASTNode ast, StringReference errorMessage){
		bool success;
		ASTNode t, t1;
		char [] op;

		t = new ASTNode();
		success = ParseLiteralVariableUnaryParenthesis(tokens, cur, t, errorMessage);

		if(success && TokenIs(tokens, cur, "^".ToCharArray())){
			op = Index(tokens, cur);
			AddToNumberReference(cur, 1d);
			t1 = new ASTNode();
			success = ParseExponentiation(tokens, cur, t1, errorMessage);
			if(success){
				t = CreateASTNode(t, t1, op);
			}
		}

		if(success){
			AssignASTNode(ast, t);
		}

		return success;
	}


	public static bool ParseLiteralVariableUnaryParenthesis(StringArrayReference tokens, NumberReference cur, ASTNode ast, StringReference errorMessage){
		bool success;
		ASTNode t;
		char [] token;

		success = true;

		token = Index(tokens, cur);
		if(IsKnownArithmeticFunction(token)){
			AddToNumberReference(cur, 1d);

			success = TokenIs(tokens, cur, "(".ToCharArray());
			if(success){
				AddToNumberReference(cur, 1d);

				t = new ASTNode();
				success = ParseAdditionOrSubtraction(tokens, cur, t, errorMessage);

				if(success){
					success = TokenIs(tokens, cur, ")".ToCharArray());
					if(success){
						ast.leaf = false;
						ast.value = token;
						ast.l = t;

						AddToNumberReference(cur, 1d);
					}else{
						errorMessage.stringx = "Epected \')\'.".ToCharArray();
					}
				}
			}else{
				errorMessage.stringx = "Epected \'(\'.".ToCharArray();
			}
		}else if(charIsLetter(token[0])){
			ast.leaf = true;
			ast.value = token;
			AddToNumberReference(cur, 1d);
		}else if(charIsNumber(token[0])){
			if(NextTokenIs(tokens, cur, "%".ToCharArray())){
				ast.leaf = false;
				ast.value = "%".ToCharArray();
				ast.l = new ASTNode();
				ast.l.leaf = true;
				ast.l.value = token;
				AddToNumberReference(cur, 1d);
				AddToNumberReference(cur, 1d);
			}else{
				ast.leaf = true;
				ast.value = token;
				AddToNumberReference(cur, 1d);
			}
		}else if(TokenIs(tokens, cur, "(".ToCharArray())){
			AddToNumberReference(cur, 1d);

			t = new ASTNode();
			success = ParseAdditionOrSubtraction(tokens, cur, t, errorMessage);

			if(success){
				ast.value = "()".ToCharArray();
				ast.l = t;
				ast.leaf = false;

				success = TokenIs(tokens, cur, ")".ToCharArray());
				if(success){
					AddToNumberReference(cur, 1d);
				}else{
					errorMessage.stringx = "Epected \')\'.".ToCharArray();
				}
			}
		}else if(TokenIs(tokens, cur, "-".ToCharArray()) || TokenIs(tokens, cur, "+".ToCharArray())){
			AddToNumberReference(cur, 1d);

			t = new ASTNode();
			success = ParseExponentiation(tokens, cur, t, errorMessage);

			if(success){
				if(arraysStringsEqual(token, "-".ToCharArray())){
					ast.value = "unary-".ToCharArray();
				}else{
					ast.value = "unary+".ToCharArray();
				}
				ast.l = t;
				ast.leaf = false;
			}
		}else{
			success = false;
			errorMessage.stringx = "Unexpected token.".ToCharArray();
		}

		return success;
	}


	public static bool IsKnownArithmeticFunction(char [] token){
		return arraysStringsEqual(token, "sqrt".ToCharArray()) || arraysStringsEqual(token, "ceil".ToCharArray()) || arraysStringsEqual(token, "floor".ToCharArray()) || arraysStringsEqual(token, "truncate".ToCharArray()) || arraysStringsEqual(token, "abs".ToCharArray()) || arraysStringsEqual(token, "log".ToCharArray()) || arraysStringsEqual(token, "ln".ToCharArray()) || arraysStringsEqual(token, "exp".ToCharArray()) || arraysStringsEqual(token, "sin".ToCharArray()) || arraysStringsEqual(token, "cos".ToCharArray()) || arraysStringsEqual(token, "tan".ToCharArray()) || arraysStringsEqual(token, "asin".ToCharArray()) || arraysStringsEqual(token, "acos".ToCharArray()) || arraysStringsEqual(token, "atan".ToCharArray());
	}


	public static bool IsKnownMatrixFunction(char [] token){
		return arraysStringsEqual(token, "trace".ToCharArray()) || arraysStringsEqual(token, "inverse".ToCharArray()) || arraysStringsEqual(token, "norm".ToCharArray()) || arraysStringsEqual(token, "minor".ToCharArray());
	}


	public static bool TokenizeBitwiseFormula(char [] f, StringArrayReference tokens, StringReference errorMessage){
		bool success;
		double i;
		LinkedListStrings ll;
		LinkedListCharacters llc;
		char [] str;

		success = true;
		ll = CreateLinkedListString();
		llc = CreateLinkedListCharacter();

		for(i = 0d; i < f.Length && success; ){
			if(f[(int)(i)] == '('){
				LinkedListAddString(ll, "(".ToCharArray());
				i = i + 1d;
			}else if(f[(int)(i)] == ')'){
				LinkedListAddString(ll, ")".ToCharArray());
				i = i + 1d;
			}else if(f[(int)(i)] == '~'){
				LinkedListAddString(ll, "~".ToCharArray());
				i = i + 1d;
			}else if(f[(int)(i)] == '&'){
				LinkedListAddString(ll, "&".ToCharArray());
				i = i + 1d;
			}else if(f[(int)(i)] == '^'){
				LinkedListAddString(ll, "^".ToCharArray());
				i = i + 1d;
			}else if(f[(int)(i)] == '|'){
				LinkedListAddString(ll, "|".ToCharArray());
				i = i + 1d;
			}else if(f[(int)(i)] == ' ' || f[(int)(i)] == '\t' || f[(int)(i)] == '\n'){
				i = i + 1d;
			}else if(charIsLetter(f[(int)(i)])){
				for(; i < f.Length && (charIsLetter(f[(int)(i)]) || charIsNumber(f[(int)(i)]) || f[(int)(i)] == '_'); i = i + 1d){
					LinkedListAddCharacter(llc, f[(int)(i)]);
				}
				str = LinkedListCharactersToArray(llc);
				LinkedListAddString(ll, str);
				FreeLinkedListCharacter(llc);
				llc = CreateLinkedListCharacter();
			}else if(charIsNumber(f[(int)(i)])){
				for(; i < f.Length && (charIsNumber(f[(int)(i)]) || f[(int)(i)] == '.'); i = i + 1d){
					LinkedListAddCharacter(llc, f[(int)(i)]);
				}
				str = LinkedListCharactersToArray(llc);
				LinkedListAddString(ll, str);
				FreeLinkedListCharacter(llc);
				llc = CreateLinkedListCharacter();
			}else if(i + 1d < f.Length){
				if(f[(int)(i)] == '>' && f[(int)(i + 1d)] == '>'){
					LinkedListAddString(ll, ">>".ToCharArray());
					i = i + 2d;
				}else if(f[(int)(i)] == '<' && f[(int)(i + 1d)] == '<'){
					LinkedListAddString(ll, "<<".ToCharArray());
					i = i + 2d;
				}else{
					success = false;
					errorMessage.stringx = "Unknown token in formula.".ToCharArray();
				}
			}else{
				success = false;
				errorMessage.stringx = "Unknown token in formula.".ToCharArray();
			}
		}

		FreeLinkedListCharacter(llc);
		if(success){
			LinkedListAddString(ll, "<end>".ToCharArray());
			tokens.stringArray = LinkedListStringsToArray(ll);
		}
		FreeLinkedListString(ll);

		return success;
	}


	public static bool ParseBitwiseTokens(StringArrayReference tokens, NumberReference cur, ASTNode ast, StringReference errorMessage){
		bool success;

		success = ParseOr(tokens, cur, ast, errorMessage);

		if(success){
			success = TokenIs(tokens, cur, "<end>".ToCharArray());

			if(!success){
				errorMessage.stringx = "Expected the end of the formula.".ToCharArray();
			}
		}

		return success;
	}


	public static bool ParseOr(StringArrayReference tokens, NumberReference cur, ASTNode ast, StringReference errorMessage){
		bool success;
		ASTNode t, t1;
		char [] op;

		t = new ASTNode();
		success = ParseXor(tokens, cur, t, errorMessage);

		for(; success && TokenIs(tokens, cur, "|".ToCharArray()); ){
			op = Index(tokens, cur);
			AddToNumberReference(cur, 1d);
			t1 = new ASTNode();
			success = ParseXor(tokens, cur, t1, errorMessage);
			if(success){
				t = CreateASTNode(t, t1, op);
			}
		}

		if(success){
			AssignASTNode(ast, t);
		}

		return success;
	}


	public static bool ParseXor(StringArrayReference tokens, NumberReference cur, ASTNode ast, StringReference errorMessage){
		bool success;
		ASTNode t, t1;
		char [] op;

		t = new ASTNode();
		success = ParseAnd(tokens, cur, t, errorMessage);

		for(; success && TokenIs(tokens, cur, "^".ToCharArray()); ){
			op = Index(tokens, cur);
			AddToNumberReference(cur, 1d);
			t1 = new ASTNode();
			success = ParseAnd(tokens, cur, t1, errorMessage);
			if(success){
				t = CreateASTNode(t, t1, op);
			}
		}

		if(success){
			AssignASTNode(ast, t);
		}

		return success;
	}


	public static bool ParseAnd(StringArrayReference tokens, NumberReference cur, ASTNode ast, StringReference errorMessage){
		bool success;
		ASTNode t, t1;
		char [] op;

		t = new ASTNode();
		success = ParseShift(tokens, cur, t, errorMessage);

		for(; success && TokenIs(tokens, cur, "&".ToCharArray()); ){
			op = Index(tokens, cur);
			AddToNumberReference(cur, 1d);
			t1 = new ASTNode();
			success = ParseShift(tokens, cur, t1, errorMessage);
			if(success){
				t = CreateASTNode(t, t1, op);
			}
		}

		if(success){
			AssignASTNode(ast, t);
		}

		return success;
	}


	public static bool ParseShift(StringArrayReference tokens, NumberReference cur, ASTNode ast, StringReference errorMessage){
		bool success;
		ASTNode t, t1;
		char [] op;

		t = new ASTNode();
		success = ParseLiteralVariableNotParenthesis(tokens, cur, t, errorMessage);

		for(; success && (TokenIs(tokens, cur, "<<".ToCharArray()) || TokenIs(tokens, cur, ">>".ToCharArray())); ){
			op = Index(tokens, cur);
			AddToNumberReference(cur, 1d);
			t1 = new ASTNode();
			success = ParseLiteralVariableNotParenthesis(tokens, cur, t1, errorMessage);
			if(success){
				t = CreateASTNode(t, t1, op);
			}
		}

		if(success){
			AssignASTNode(ast, t);
		}

		return success;
	}


	public static bool ParseLiteralVariableNotParenthesis(StringArrayReference tokens, NumberReference cur, ASTNode ast, StringReference errorMessage){
		bool success;
		ASTNode t;
		char [] token;

		success = true;

		token = Index(tokens, cur);
		if(charIsLetter(token[0]) || charIsNumber(token[0])){
			ast.leaf = true;
			ast.value = token;
			AddToNumberReference(cur, 1d);
		}else if(TokenIs(tokens, cur, "(".ToCharArray())){
			AddToNumberReference(cur, 1d);

			t = new ASTNode();
			success = ParseOr(tokens, cur, t, errorMessage);

			if(success){
				ast.value = "()".ToCharArray();
				ast.l = t;
				ast.leaf = false;

				success = TokenIs(tokens, cur, ")".ToCharArray());
				if(success){
					AddToNumberReference(cur, 1d);
				}else{
					errorMessage.stringx = "Epected \')\'.".ToCharArray();
				}
			}
		}else if(TokenIs(tokens, cur, "~".ToCharArray())){
			AddToNumberReference(cur, 1d);

			t = new ASTNode();
			success = ParseLiteralVariableNotParenthesis(tokens, cur, t, errorMessage);

			if(success){
				ast.value = "~".ToCharArray();
				ast.l = t;
				ast.leaf = false;
			}
		}else{
			success = false;
			errorMessage.stringx = "Unexpected token.".ToCharArray();
		}

		return success;
	}


	public static void ASTToTFormFunctions(ASTNode ast, StringReference tf, char [] prefix, char [] postfix, char [] tprefix, bool parenthesis, bool semicolon, bool wrappedNumber, char [] target){
		BooleanArrayReference t;
		NumberReference assignedT;

		t = CreateBooleanArrayReferenceLengthValue(0d, false);
		assignedT = CreateNumberReference(0d);
		tf.stringx = new char [0];

		ASTToTFormFunctionsInner(ast, tf, t, assignedT, prefix, postfix, tprefix, parenthesis, semicolon, wrappedNumber, target, 0d);
	}


	public static void ASTToTFormFunctionsInner(ASTNode ast, StringReference tf, BooleanArrayReference ts, NumberReference assignedT, char [] prefix, char [] postfix, char [] tprefix, bool parenthesis, bool semicolon, bool wrappedNumber, char [] target, double level){
		double tl, tr, t;
		char [] functionName, value, numberString;

		tl = 0d;
		tr = 0d;

		if(!ast.leaf){
			if(arraysStringsEqual(ast.value, "+".ToCharArray()) || arraysStringsEqual(ast.value, "-".ToCharArray()) || arraysStringsEqual(ast.value, "mod".ToCharArray()) || arraysStringsEqual(ast.value, "*".ToCharArray()) || arraysStringsEqual(ast.value, "/".ToCharArray()) || arraysStringsEqual(ast.value, "^".ToCharArray())){
				if(!ast.l.leaf){
					ASTToTFormFunctionsInner(ast.l, tf, ts, assignedT, prefix, postfix, tprefix, parenthesis, semicolon, wrappedNumber, target, level + 1d);
					tl = assignedT.numberValue;
				}

				if(!ast.r.leaf){
					ASTToTFormFunctionsInner(ast.r, tf, ts, assignedT, prefix, postfix, tprefix, parenthesis, semicolon, wrappedNumber, target, level + 1d);
					tr = assignedT.numberValue;
				}

				if(!ast.l.leaf){
					FreeTVariable(ts, tl);
				}

				if(!ast.r.leaf){
					FreeTVariable(ts, tr);
				}

				functionName = ArithmeticBinarySymbolToFunctionName(ast.value);

				tf.stringx = AppendString(tf.stringx, prefix);
				tf.stringx = AppendString(tf.stringx, functionName);
				tf.stringx = AppendString(tf.stringx, postfix);
				if(parenthesis){
					tf.stringx = AppendString(tf.stringx, "(".ToCharArray());
				}else{
					tf.stringx = AppendString(tf.stringx, " ".ToCharArray());
				}

				if(level > 0d || target.Length == 0d){
					t = AllocateTVariable(ts);
					assignedT.numberValue = t;
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					tf.stringx = AppendString(tf.stringx, tprefix);
					numberString = CreateStringDecimalFromNumber(t);
					tf.stringx = AppendString(tf.stringx, numberString);
				}else{
					tf.stringx = AppendString(tf.stringx, target);
				}

				tf.stringx = AppendString(tf.stringx, ", ".ToCharArray());

				if(ast.l.leaf){
					if(charIsNumber(ast.l.value[0])){
						value = new char [0];
						if(wrappedNumber){
							value = AppendString(value, prefix);
							value = AppendString(value, "Number".ToCharArray());
							value = AppendString(value, postfix);
							value = AppendString(value, "(".ToCharArray());
						}
						value = AppendString(value, ast.l.value);
						if(wrappedNumber){
							value = AppendString(value, ")".ToCharArray());
						}
					}else{
						value = ast.l.value;
					}
					tf.stringx = AppendString(tf.stringx, value);
				}else{
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					tf.stringx = AppendString(tf.stringx, tprefix);
					numberString = CreateStringDecimalFromNumber(tl);
					tf.stringx = AppendString(tf.stringx, numberString);
				}
				tf.stringx = AppendString(tf.stringx, ", ".ToCharArray());
				if(ast.r.leaf){
					if(charIsNumber(ast.r.value[0])){
						value = new char [0];
						if(wrappedNumber){
							value = AppendString(value, prefix);
							value = AppendString(value, "Number".ToCharArray());
							value = AppendString(value, postfix);
							value = AppendString(value, "(".ToCharArray());
						}
						value = AppendString(value, ast.r.value);
						if(wrappedNumber){
							value = AppendString(value, ")".ToCharArray());
						}
					}else{
						value = ast.r.value;
					}
					tf.stringx = AppendString(tf.stringx, value);
				}else{
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					tf.stringx = AppendString(tf.stringx, tprefix);
					numberString = CreateStringDecimalFromNumber(tr);
					tf.stringx = AppendString(tf.stringx, numberString);
				}
				if(parenthesis){
					tf.stringx = AppendString(tf.stringx, ")".ToCharArray());
				}else{
					tf.stringx = AppendString(tf.stringx, " ".ToCharArray());
				}
				if(semicolon){
					tf.stringx = AppendString(tf.stringx, ";".ToCharArray());
				}
				tf.stringx = AppendString(tf.stringx, "\n".ToCharArray());
			}else if(arraysStringsEqual(ast.value, "()".ToCharArray())){
				ASTToTFormFunctionsInner(ast.l, tf, ts, assignedT, prefix, postfix, tprefix, parenthesis, semicolon, wrappedNumber, target, level + 1d);
			}else if(arraysStringsEqual(ast.value, "unary-".ToCharArray()) || arraysStringsEqual(ast.value, "unary+".ToCharArray())){
				if(!ast.l.leaf){
					ASTToTFormFunctionsInner(ast.l, tf, ts, assignedT, prefix, postfix, tprefix, parenthesis, semicolon, wrappedNumber, target, level + 1d);
					tl = assignedT.numberValue;
					FreeTVariable(ts, tl);
				}

				functionName = ArithmeticBinarySymbolToFunctionName(ast.value);

				tf.stringx = AppendString(tf.stringx, prefix);
				tf.stringx = AppendString(tf.stringx, functionName);
				tf.stringx = AppendString(tf.stringx, postfix);
				if(parenthesis){
					tf.stringx = AppendString(tf.stringx, "(".ToCharArray());
				}else{
					tf.stringx = AppendString(tf.stringx, " ".ToCharArray());
				}

				if(level > 0d || target.Length == 0d){
					t = AllocateTVariable(ts);
					assignedT.numberValue = t;
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					tf.stringx = AppendString(tf.stringx, tprefix);
					numberString = CreateStringDecimalFromNumber(t);
					tf.stringx = AppendString(tf.stringx, numberString);
				}else{
					tf.stringx = AppendString(tf.stringx, target);
				}

				tf.stringx = AppendString(tf.stringx, ", ".ToCharArray());

				if(ast.l.leaf){
					tf.stringx = AppendString(tf.stringx, ast.l.value);
				}else{
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					tf.stringx = AppendString(tf.stringx, tprefix);
					numberString = CreateStringDecimalFromNumber(tl);
					tf.stringx = AppendString(tf.stringx, numberString);
				}
				if(parenthesis){
					tf.stringx = AppendString(tf.stringx, ")".ToCharArray());
				}else{
					tf.stringx = AppendString(tf.stringx, " ".ToCharArray());
				}
				if(semicolon){
					tf.stringx = AppendString(tf.stringx, ";".ToCharArray());
				}
				tf.stringx = AppendString(tf.stringx, "\n".ToCharArray());
			}else if(IsKnownArithmeticFunction(ast.value)){
				if(!ast.l.leaf){
					ASTToTFormFunctionsInner(ast.l, tf, ts, assignedT, prefix, postfix, tprefix, parenthesis, semicolon, wrappedNumber, target, level + 1d);
					tl = assignedT.numberValue;
					FreeTVariable(ts, tl);
				}

				functionName = arraysCopyString(ast.value);
				functionName[0] = charToUpperCase(functionName[0]);

				tf.stringx = AppendString(tf.stringx, prefix);
				tf.stringx = AppendString(tf.stringx, functionName);
				tf.stringx = AppendString(tf.stringx, postfix);
				if(parenthesis){
					tf.stringx = AppendString(tf.stringx, "(".ToCharArray());
				}else{
					tf.stringx = AppendString(tf.stringx, " ".ToCharArray());
				}

				if(level > 0d || target.Length == 0d){
					t = AllocateTVariable(ts);
					assignedT.numberValue = t;
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					tf.stringx = AppendString(tf.stringx, tprefix);
					numberString = CreateStringDecimalFromNumber(t);
					tf.stringx = AppendString(tf.stringx, numberString);
				}else{
					tf.stringx = AppendString(tf.stringx, target);
				}

				tf.stringx = AppendString(tf.stringx, ", ".ToCharArray());

				if(ast.l.leaf){
					tf.stringx = AppendString(tf.stringx, ast.l.value);
				}else{
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					tf.stringx = AppendString(tf.stringx, tprefix);
					numberString = CreateStringDecimalFromNumber(tl);
					tf.stringx = AppendString(tf.stringx, numberString);
				}
				if(parenthesis){
					tf.stringx = AppendString(tf.stringx, ")".ToCharArray());
				}else{
					tf.stringx = AppendString(tf.stringx, " ".ToCharArray());
				}
				if(semicolon){
					tf.stringx = AppendString(tf.stringx, ";".ToCharArray());
				}
				tf.stringx = AppendString(tf.stringx, "\n".ToCharArray());
			}else if(arraysStringsEqual(ast.value, "%".ToCharArray())){
				tf.stringx = AppendString(tf.stringx, prefix);
				tf.stringx = AppendString(tf.stringx, "Percentage".ToCharArray());
				tf.stringx = AppendString(tf.stringx, postfix);
				if(parenthesis){
					tf.stringx = AppendString(tf.stringx, "(".ToCharArray());
				}else{
					tf.stringx = AppendString(tf.stringx, " ".ToCharArray());
				}

				if(level > 0d || target.Length == 0d){
					t = AllocateTVariable(ts);
					assignedT.numberValue = t;
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					tf.stringx = AppendString(tf.stringx, tprefix);
					numberString = CreateStringDecimalFromNumber(t);
					tf.stringx = AppendString(tf.stringx, numberString);
				}else{
					tf.stringx = AppendString(tf.stringx, target);
				}

				tf.stringx = AppendString(tf.stringx, ", ".ToCharArray());

				tf.stringx = AppendString(tf.stringx, ast.l.value);
				if(parenthesis){
					tf.stringx = AppendString(tf.stringx, ")".ToCharArray());
				}else{
					tf.stringx = AppendString(tf.stringx, " ".ToCharArray());
				}
				if(semicolon){
					tf.stringx = AppendString(tf.stringx, ";".ToCharArray());
				}
				tf.stringx = AppendString(tf.stringx, "\n".ToCharArray());
			}else{
				tf.stringx = AppendString(tf.stringx, "<failed>".ToCharArray());
			}
		}else{
			tf.stringx = AppendString(tf.stringx, "<failed>".ToCharArray());
		}
	}


	public static char [] ArithmeticBinarySymbolToFunctionName(char [] value){
		char [] f;

		f = "Unknown".ToCharArray();

		if(arraysStringsEqual(value, "+".ToCharArray())){
			f = "Add".ToCharArray();
		}
		if(arraysStringsEqual(value, "-".ToCharArray())){
			f = "Sub".ToCharArray();
		}
		if(arraysStringsEqual(value, "mod".ToCharArray())){
			f = "Mod".ToCharArray();
		}
		if(arraysStringsEqual(value, "*".ToCharArray())){
			f = "Mul".ToCharArray();
		}
		if(arraysStringsEqual(value, "/".ToCharArray())){
			f = "Div".ToCharArray();
		}
		if(arraysStringsEqual(value, "^".ToCharArray())){
			f = "Pow".ToCharArray();
		}
		if(arraysStringsEqual(value, "unary-".ToCharArray())){
			f = "Negate".ToCharArray();
		}
		if(arraysStringsEqual(value, "unary+".ToCharArray())){
			f = "Positive".ToCharArray();
		}

		return f;
	}


	public static bool ArithmeticFormulaToTFormFunctions(char [] f, char [] prefix, char [] postfix, char [] tprefix, bool parenthesis, bool semicolon, bool wrappedNumber, char [] target, StringReference result, StringReference message){
		StringArrayReference tokens;
		bool success;
		ASTNode ast;
		NumberReference pos;

		tokens = new StringArrayReference();
		success = TokenizeArithmeticFormula(f, tokens, message);

		if(success){
			/* Parse*/
			ast = new ASTNode();
			pos = CreateNumberReference(0d);
			success = ParseArithmeticTokens(tokens, pos, ast, message);

			if(success){
				ASTToTFormFunctions(ast, result, prefix, postfix, tprefix, parenthesis, semicolon, wrappedNumber, target);
			}
		}

		return success;
	}


	public static double AllocateTVariable(BooleanArrayReference ts){
		double i, newT;
		bool found;
		bool [] newArray;

		newT = 0d;
		found = false;

		for(i = 0d; i < ts.booleanArray.Length && !found; i = i + 1d){
			if(!ts.booleanArray[(int)(i)]){
				found = true;
				newT = i;
			}
		}

		if(found){
		}else{
			newT = i;
			newArray = new bool [(int)(ts.booleanArray.Length + 1d)];
			for(i = 0d; i < ts.booleanArray.Length; i = i + 1d){
				newArray[(int)(i)] = ts.booleanArray[(int)(i)];
			}
			newArray[(int)(newT)] = false;
			delete(ts.booleanArray);
			ts.booleanArray = newArray;
		}

		/* Mark as allocated:*/
		ts.booleanArray[(int)(newT)] = true;

		return newT;
	}


	public static void FreeTVariable(BooleanArrayReference ts, double t){
		ts.booleanArray[(int)(t)] = false;
	}


	public static char [] ASTToBooleanFormula(ASTNode ast){
		char [] f, p1, p2;

		if(ast.leaf){
			f = ast.value;
		}else if(arraysStringsEqual(ast.value, "!".ToCharArray())){
			f = new char [0];
			p1 = ASTToBooleanFormula(ast.l);
			f = AppendString(f, "!".ToCharArray());
			f = AppendString(f, p1);
		}else if(arraysStringsEqual(ast.value, "()".ToCharArray())){
			f = new char [0];
			p1 = ASTToBooleanFormula(ast.l);
			f = AppendString(f, "(".ToCharArray());
			f = AppendString(f, p1);
			f = AppendString(f, ")".ToCharArray());
		}else if(arraysStringsEqual(ast.value, "&".ToCharArray()) || arraysStringsEqual(ast.value, "|".ToCharArray()) || arraysStringsEqual(ast.value, "^".ToCharArray()) || arraysStringsEqual(ast.value, "=".ToCharArray()) || arraysStringsEqual(ast.value, "!=".ToCharArray()) || arraysStringsEqual(ast.value, "<".ToCharArray()) || arraysStringsEqual(ast.value, "<=".ToCharArray()) || arraysStringsEqual(ast.value, ">".ToCharArray()) || arraysStringsEqual(ast.value, ">=".ToCharArray()) || arraysStringsEqual(ast.value, "->".ToCharArray()) || arraysStringsEqual(ast.value, "<->".ToCharArray())){
			f = new char [0];
			p1 = ASTToBooleanFormula(ast.l);
			p2 = ASTToBooleanFormula(ast.r);
			f = AppendString(f, p1);
			f = AppendString(f, " ".ToCharArray());
			f = AppendString(f, ast.value);
			f = AppendString(f, " ".ToCharArray());
			f = AppendString(f, p2);
		}else{
			f = "<failed>".ToCharArray();
		}

		return f;
	}


	public static void BooleanASTToTForm(ASTNode ast, StringReference tf, NumberReference t){
		double tl, tr;
		StringReference numberString;
		numberString = new StringReference();

		tl = 0d;
		tr = 0d;

		if(!ast.leaf){
			if(arraysStringsEqual(ast.value, "&".ToCharArray()) || arraysStringsEqual(ast.value, "|".ToCharArray()) || arraysStringsEqual(ast.value, "^".ToCharArray()) || arraysStringsEqual(ast.value, "=".ToCharArray()) || arraysStringsEqual(ast.value, "!=".ToCharArray()) || arraysStringsEqual(ast.value, "<".ToCharArray()) || arraysStringsEqual(ast.value, "<=".ToCharArray()) || arraysStringsEqual(ast.value, ">".ToCharArray()) || arraysStringsEqual(ast.value, ">=".ToCharArray()) || arraysStringsEqual(ast.value, "->".ToCharArray()) || arraysStringsEqual(ast.value, "<->".ToCharArray())){
				if(!ast.l.leaf){
					BooleanASTToTForm(ast.l, tf, t);
					tl = t.numberValue;
				}

				if(!ast.r.leaf){
					BooleanASTToTForm(ast.r, tf, t);
					tr = t.numberValue;
				}

				t.numberValue = t.numberValue + 1d;
				tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
				CreateStringFromNumberWithCheck(t.numberValue, 10d, numberString);
				tf.stringx = AppendString(tf.stringx, numberString.stringx);
				tf.stringx = AppendString(tf.stringx, " = ".ToCharArray());
				if(ast.l.leaf){
					tf.stringx = AppendString(tf.stringx, ast.l.value);
				}else{
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					CreateStringFromNumberWithCheck(tl, 10d, numberString);
					tf.stringx = AppendString(tf.stringx, numberString.stringx);
				}
				tf.stringx = AppendString(tf.stringx, " ".ToCharArray());
				tf.stringx = AppendString(tf.stringx, ast.value);
				tf.stringx = AppendString(tf.stringx, " ".ToCharArray());
				if(ast.r.leaf){
					tf.stringx = AppendString(tf.stringx, ast.r.value);
				}else{
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					CreateStringFromNumberWithCheck(tr, 10d, numberString);
					tf.stringx = AppendString(tf.stringx, numberString.stringx);
				}
				tf.stringx = AppendString(tf.stringx, ";\n".ToCharArray());
			}else if(arraysStringsEqual(ast.value, "()".ToCharArray())){
				BooleanASTToTForm(ast.l, tf, t);
				tl = t.numberValue;

				t.numberValue = t.numberValue + 1d;
				tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
				CreateStringFromNumberWithCheck(t.numberValue, 10d, numberString);
				tf.stringx = AppendString(tf.stringx, numberString.stringx);
				tf.stringx = AppendString(tf.stringx, " = ".ToCharArray());
				tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
				CreateStringFromNumberWithCheck(tl, 10d, numberString);
				tf.stringx = AppendString(tf.stringx, numberString.stringx);
				tf.stringx = AppendString(tf.stringx, ";\n".ToCharArray());
			}else if(arraysStringsEqual(ast.value, "!".ToCharArray())){
				if(!ast.l.leaf){
					BooleanASTToTForm(ast.l, tf, t);
					tl = t.numberValue;
				}

				t.numberValue = t.numberValue + 1d;
				tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
				CreateStringFromNumberWithCheck(t.numberValue, 10d, numberString);
				tf.stringx = AppendString(tf.stringx, numberString.stringx);
				tf.stringx = AppendString(tf.stringx, " = ".ToCharArray());
				tf.stringx = AppendString(tf.stringx, "!".ToCharArray());
				if(ast.l.leaf){
					tf.stringx = AppendString(tf.stringx, ast.l.value);
				}else{
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					CreateStringFromNumberWithCheck(tl, 10d, numberString);
					tf.stringx = AppendString(tf.stringx, numberString.stringx);
				}
				tf.stringx = AppendString(tf.stringx, ";\n".ToCharArray());
			}else{
				tf.stringx = AppendString(tf.stringx, "<failed>".ToCharArray());
			}
		}
	}


	public static char [] BooleanFormulaToTForm(char [] f){
		char [] af;
		StringArrayReference tokens;
		bool success;
		StringReference errorMessage, tf;
		ASTNode ast;
		NumberReference pos, t;

		tokens = new StringArrayReference();
		errorMessage = new StringReference();
		success = TokenizeBooleanFormula(f, tokens, errorMessage);

		if(success){
			/* Parse*/
			ast = new ASTNode();
			pos = CreateNumberReference(0d);
			success = ParseBooleanTokens(tokens, pos, ast, errorMessage);

			if(success){
				tf = CreateStringReference("".ToCharArray());
				t = CreateNumberReference(0d);
				BooleanASTToTForm(ast, tf, t);
				af = tf.stringx;
			}else{
				af = errorMessage.stringx;
			}
		}else{
			af = errorMessage.stringx;
		}

		return af;
	}


	public static void BitwiseASTToTFormFunctions(ASTNode ast, StringReference tf, char [] prefix, char [] postfix, char [] tprefix, bool parenthesis, bool semicolon, bool wrappedNumber, char [] target){
		BooleanArrayReference t;
		NumberReference assignedT;

		t = CreateBooleanArrayReferenceLengthValue(0d, false);
		assignedT = CreateNumberReference(0d);
		tf.stringx = new char [0];

		BitwiseASTToTFormFunctionsInner(ast, tf, t, assignedT, prefix, postfix, tprefix, parenthesis, semicolon, wrappedNumber, target, 0d);
	}


	public static void BitwiseASTToTFormFunctionsInner(ASTNode ast, StringReference tf, BooleanArrayReference ts, NumberReference assignedT, char [] prefix, char [] postfix, char [] tprefix, bool parenthesis, bool semicolon, bool wrappedNumber, char [] target, double level){
		double tl, tr, t;
		char [] functionName, value, numberString;

		tl = 0d;
		tr = 0d;
		t = 0d;

		if(!ast.leaf){
			if(arraysStringsEqual(ast.value, "&".ToCharArray()) || arraysStringsEqual(ast.value, "|".ToCharArray()) || arraysStringsEqual(ast.value, "^".ToCharArray()) || arraysStringsEqual(ast.value, "<<".ToCharArray()) || arraysStringsEqual(ast.value, ">>".ToCharArray())){
				if(!ast.l.leaf){
					BitwiseASTToTFormFunctionsInner(ast.l, tf, ts, assignedT, prefix, postfix, tprefix, parenthesis, semicolon, wrappedNumber, target, level + 1d);
					tl = assignedT.numberValue;
				}

				if(!ast.r.leaf){
					BitwiseASTToTFormFunctionsInner(ast.r, tf, ts, assignedT, prefix, postfix, tprefix, parenthesis, semicolon, wrappedNumber, target, level + 1d);
					tr = assignedT.numberValue;
				}

				if(!ast.l.leaf){
					FreeTVariable(ts, tl);
				}

				if(!ast.r.leaf){
					FreeTVariable(ts, tr);
				}

				functionName = BitwiseBinarySymbolToFunctionName(ast.value);

				tf.stringx = AppendString(tf.stringx, prefix);
				tf.stringx = AppendString(tf.stringx, functionName);
				tf.stringx = AppendString(tf.stringx, postfix);
				if(parenthesis){
					tf.stringx = AppendString(tf.stringx, "(".ToCharArray());
				}else{
					tf.stringx = AppendString(tf.stringx, " ".ToCharArray());
				}

				if(level > 0d || target.Length == 0d){
					t = AllocateTVariable(ts);
					assignedT.numberValue = t;
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					tf.stringx = AppendString(tf.stringx, tprefix);
					numberString = CreateStringDecimalFromNumber(t);
					tf.stringx = AppendString(tf.stringx, numberString);
				}else{
					tf.stringx = AppendString(tf.stringx, target);
				}

				tf.stringx = AppendString(tf.stringx, ", ".ToCharArray());

				if(ast.l.leaf){
					if(charIsNumber(ast.l.value[0])){
						value = new char [0];
						value = AppendString(value, prefix);
						if(wrappedNumber){
							value = AppendString(value, "Number".ToCharArray());
							value = AppendString(value, postfix);
							value = AppendString(value, "(".ToCharArray());
						}
						value = AppendString(value, ast.l.value);
						if(wrappedNumber){
							value = AppendString(value, ")".ToCharArray());
						}
					}else{
						value = ast.l.value;
					}
					tf.stringx = AppendString(tf.stringx, value);
				}else{
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					tf.stringx = AppendString(tf.stringx, tprefix);
					numberString = CreateStringDecimalFromNumber(tl);
					tf.stringx = AppendString(tf.stringx, numberString);
				}
				tf.stringx = AppendString(tf.stringx, ", ".ToCharArray());
				if(ast.r.leaf){
					if(charIsNumber(ast.r.value[0])){
						value = new char [0];
						value = AppendString(value, prefix);
						if(wrappedNumber){
							value = AppendString(value, "Number".ToCharArray());
							value = AppendString(value, postfix);
							value = AppendString(value, "(".ToCharArray());
						}
						value = AppendString(value, ast.r.value);
						if(wrappedNumber){
							value = AppendString(value, ")".ToCharArray());
						}
					}else{
						value = ast.r.value;
					}
					tf.stringx = AppendString(tf.stringx, value);
				}else{
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					tf.stringx = AppendString(tf.stringx, tprefix);
					numberString = CreateStringDecimalFromNumber(tr);
					tf.stringx = AppendString(tf.stringx, numberString);
				}
				if(parenthesis){
					tf.stringx = AppendString(tf.stringx, ")".ToCharArray());
				}else{
					tf.stringx = AppendString(tf.stringx, " ".ToCharArray());
				}
				if(semicolon){
					tf.stringx = AppendString(tf.stringx, ";".ToCharArray());
				}
				tf.stringx = AppendString(tf.stringx, "\n".ToCharArray());
			}else if(arraysStringsEqual(ast.value, "()".ToCharArray())){
				BitwiseASTToTFormFunctionsInner(ast.l, tf, ts, assignedT, prefix, postfix, tprefix, parenthesis, semicolon, wrappedNumber, target, level + 1d);
			}else if(arraysStringsEqual(ast.value, "~".ToCharArray())){
				if(!ast.l.leaf){
					BitwiseASTToTFormFunctionsInner(ast.l, tf, ts, assignedT, prefix, postfix, tprefix, parenthesis, semicolon, wrappedNumber, target, level + 1d);
					tl = assignedT.numberValue;
					FreeTVariable(ts, tl);
				}

				functionName = BitwiseBinarySymbolToFunctionName(ast.value);

				tf.stringx = AppendString(tf.stringx, prefix);
				tf.stringx = AppendString(tf.stringx, functionName);
				tf.stringx = AppendString(tf.stringx, postfix);
				if(parenthesis){
					tf.stringx = AppendString(tf.stringx, "(".ToCharArray());
				}else{
					tf.stringx = AppendString(tf.stringx, " ".ToCharArray());
				}

				if(level > 0d || target.Length == 0d){
					t = AllocateTVariable(ts);
					assignedT.numberValue = t;
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					tf.stringx = AppendString(tf.stringx, tprefix);
					numberString = CreateStringDecimalFromNumber(t);
					tf.stringx = AppendString(tf.stringx, numberString);
				}else{
					tf.stringx = AppendString(tf.stringx, target);
				}

				tf.stringx = AppendString(tf.stringx, ", ".ToCharArray());

				if(ast.l.leaf){
					tf.stringx = AppendString(tf.stringx, ast.l.value);
				}else{
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					tf.stringx = AppendString(tf.stringx, tprefix);
					numberString = CreateStringDecimalFromNumber(tl);
					tf.stringx = AppendString(tf.stringx, numberString);
				}
				if(parenthesis){
					tf.stringx = AppendString(tf.stringx, ")".ToCharArray());
				}else{
					tf.stringx = AppendString(tf.stringx, " ".ToCharArray());
				}
				if(semicolon){
					tf.stringx = AppendString(tf.stringx, ";".ToCharArray());
				}
				tf.stringx = AppendString(tf.stringx, "\n".ToCharArray());
			}else{
				tf.stringx = AppendString(tf.stringx, "<failed>".ToCharArray());
			}
		}else{
			tf.stringx = AppendString(tf.stringx, "<failed>".ToCharArray());
		}
	}


	public static char [] BitwiseBinarySymbolToFunctionName(char [] value){
		char [] f;

		f = "Unknown".ToCharArray();

		if(arraysStringsEqual(value, "~".ToCharArray())){
			f = "Not".ToCharArray();
		}
		if(arraysStringsEqual(value, "&".ToCharArray())){
			f = "And".ToCharArray();
		}
		if(arraysStringsEqual(value, "|".ToCharArray())){
			f = "Or".ToCharArray();
		}
		if(arraysStringsEqual(value, "^".ToCharArray())){
			f = "Xor".ToCharArray();
		}
		if(arraysStringsEqual(value, "<<".ToCharArray())){
			f = "ShiftLeft".ToCharArray();
		}
		if(arraysStringsEqual(value, ">>".ToCharArray())){
			f = "ShiftRight".ToCharArray();
		}

		return f;
	}


	public static bool BitwiseFormulaToTFormFunctions(char [] f, char [] prefix, char [] postfix, char [] tprefix, bool parenthesis, bool semicolon, bool wrappedNumber, char [] target, StringReference tf, StringReference message){
		StringArrayReference tokens;
		bool success;
		ASTNode ast;
		NumberReference pos;

		tokens = new StringArrayReference();
		success = TokenizeBitwiseFormula(f, tokens, message);

		if(success){
			/* Parse*/
			ast = new ASTNode();
			pos = CreateNumberReference(0d);
			success = ParseBitwiseTokens(tokens, pos, ast, message);

			if(success){
				BitwiseASTToTFormFunctions(ast, tf, prefix, postfix, tprefix, parenthesis, semicolon, wrappedNumber, target);
			}
		}

		return success;
	}


	public static void AddToNumberReference(NumberReference cur, double val){
		cur.numberValue = cur.numberValue + val;
	}


	public static bool TokenIs(StringArrayReference tokens, NumberReference cur, char [] s){
		return arraysStringsEqual(Index(tokens, cur), s);
	}


	public static bool NextTokenIs(StringArrayReference tokens, NumberReference cur, char [] s){
		return arraysStringsEqual(NextIndex(tokens, cur), s);
	}


	public static char [] NextIndex(StringArrayReference stringArrayReference, NumberReference index){
		return stringArrayReference.stringArray[(int)(index.numberValue + 1d)].stringx;
	}


	public static char [] Index(StringArrayReference stringArrayReference, NumberReference index){
		return stringArrayReference.stringArray[(int)(index.numberValue)].stringx;
	}


	public static void AssignASTNode(ASTNode x, ASTNode a){
		x.leaf = a.leaf;
		x.l = a.l;
		x.r = a.r;
		x.value = a.value;
	}


	public static ASTNode CreateASTNode(ASTNode l, ASTNode r, char [] op){
		ASTNode t;

		t = new ASTNode();
		t.value = op;
		t.l = l;
		t.r = r;
		t.leaf = false;

		return t;
	}


	public static char [] ASTToArithmeticFormula(ASTNode ast){
		char [] f, p1, p2;

		if(ast.leaf){
			f = ast.value;
		}else if(arraysStringsEqual(ast.value, "unary-".ToCharArray())){
			f = new char [0];
			p1 = ASTToArithmeticFormula(ast.l);
			f = AppendString(f, "-".ToCharArray());
			f = AppendString(f, p1);
		}else if(arraysStringsEqual(ast.value, "unary+".ToCharArray())){
			f = new char [0];
			p1 = ASTToArithmeticFormula(ast.l);
			f = AppendString(f, "+".ToCharArray());
			f = AppendString(f, p1);
		}else if(arraysStringsEqual(ast.value, "%".ToCharArray())){
			f = new char [0];
			f = AppendString(f, ast.l.value);
			f = AppendString(f, "%".ToCharArray());
		}else if(IsKnownArithmeticFunction(ast.value)){
			f = new char [0];
			p1 = ASTToArithmeticFormula(ast.l);
			f = AppendString(f, ast.value);
			f = AppendString(f, "(".ToCharArray());
			f = AppendString(f, p1);
			f = AppendString(f, ")".ToCharArray());
		}else if(arraysStringsEqual(ast.value, "()".ToCharArray())){
			f = new char [0];
			p1 = ASTToArithmeticFormula(ast.l);
			f = AppendString(f, "(".ToCharArray());
			f = AppendString(f, p1);
			f = AppendString(f, ")".ToCharArray());
		}else if(arraysStringsEqual(ast.value, "+".ToCharArray()) || arraysStringsEqual(ast.value, "-".ToCharArray()) || arraysStringsEqual(ast.value, "mod".ToCharArray())){
			f = new char [0];
			p1 = ASTToArithmeticFormula(ast.l);
			p2 = ASTToArithmeticFormula(ast.r);
			f = AppendString(f, p1);
			f = AppendString(f, " ".ToCharArray());
			f = AppendString(f, ast.value);
			f = AppendString(f, " ".ToCharArray());
			f = AppendString(f, p2);
		}else if(arraysStringsEqual(ast.value, "*".ToCharArray()) || arraysStringsEqual(ast.value, "/".ToCharArray()) || arraysStringsEqual(ast.value, "^".ToCharArray())){
			f = new char [0];
			p1 = ASTToArithmeticFormula(ast.l);
			p2 = ASTToArithmeticFormula(ast.r);
			f = AppendString(f, p1);
			f = AppendString(f, ast.value);
			f = AppendString(f, p2);
		}else{
			f = "<failed>".ToCharArray();
		}

		return f;
	}


	public static void ArithmeticASTToTForm(ASTNode ast, StringReference tf, NumberReference t){
		double tl, tr;
		StringReference numberString;
		numberString = new StringReference();

		tl = 0d;
		tr = 0d;

		if(!ast.leaf){
			if(arraysStringsEqual(ast.value, "+".ToCharArray()) || arraysStringsEqual(ast.value, "-".ToCharArray()) || arraysStringsEqual(ast.value, "mod".ToCharArray()) || arraysStringsEqual(ast.value, "*".ToCharArray()) || arraysStringsEqual(ast.value, "/".ToCharArray()) || arraysStringsEqual(ast.value, "^".ToCharArray())){
				if(!ast.l.leaf){
					ArithmeticASTToTForm(ast.l, tf, t);
					tl = t.numberValue;
				}

				if(!ast.r.leaf){
					ArithmeticASTToTForm(ast.r, tf, t);
					tr = t.numberValue;
				}

				t.numberValue = t.numberValue + 1d;
				tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
				CreateStringFromNumberWithCheck(t.numberValue, 10d, numberString);
				tf.stringx = AppendString(tf.stringx, numberString.stringx);
				tf.stringx = AppendString(tf.stringx, " = ".ToCharArray());
				if(ast.l.leaf){
					tf.stringx = AppendString(tf.stringx, ast.l.value);
				}else{
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					CreateStringFromNumberWithCheck(tl, 10d, numberString);
					tf.stringx = AppendString(tf.stringx, numberString.stringx);
				}
				tf.stringx = AppendString(tf.stringx, " ".ToCharArray());
				tf.stringx = AppendString(tf.stringx, ast.value);
				tf.stringx = AppendString(tf.stringx, " ".ToCharArray());
				if(ast.r.leaf){
					tf.stringx = AppendString(tf.stringx, ast.r.value);
				}else{
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					CreateStringFromNumberWithCheck(tr, 10d, numberString);
					tf.stringx = AppendString(tf.stringx, numberString.stringx);
				}
				tf.stringx = AppendString(tf.stringx, ";\n".ToCharArray());
			}else if(arraysStringsEqual(ast.value, "()".ToCharArray())){
				ArithmeticASTToTForm(ast.l, tf, t);
			}else if(arraysStringsEqual(ast.value, "unary-".ToCharArray()) || arraysStringsEqual(ast.value, "unary+".ToCharArray())){
				if(!ast.l.leaf){
					ArithmeticASTToTForm(ast.l, tf, t);
					tl = t.numberValue;
				}

				t.numberValue = t.numberValue + 1d;
				tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
				CreateStringFromNumberWithCheck(t.numberValue, 10d, numberString);
				tf.stringx = AppendString(tf.stringx, numberString.stringx);
				tf.stringx = AppendString(tf.stringx, " = ".ToCharArray());
				if(arraysStringsEqual(ast.value, "unary-".ToCharArray())){
					tf.stringx = AppendString(tf.stringx, "-".ToCharArray());
				}else{
					tf.stringx = AppendString(tf.stringx, "+".ToCharArray());
				}
				if(ast.l.leaf){
					tf.stringx = AppendString(tf.stringx, ast.l.value);
				}else{
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					CreateStringFromNumberWithCheck(tl, 10d, numberString);
					tf.stringx = AppendString(tf.stringx, numberString.stringx);
				}
				tf.stringx = AppendString(tf.stringx, ";\n".ToCharArray());
			}else if(IsKnownArithmeticFunction(ast.value)){
				if(!ast.l.leaf){
					ArithmeticASTToTForm(ast.l, tf, t);
					tl = t.numberValue;
				}

				t.numberValue = t.numberValue + 1d;
				tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
				CreateStringFromNumberWithCheck(t.numberValue, 10d, numberString);
				tf.stringx = AppendString(tf.stringx, numberString.stringx);
				tf.stringx = AppendString(tf.stringx, " = ".ToCharArray());
				tf.stringx = AppendString(tf.stringx, ast.value);
				tf.stringx = AppendString(tf.stringx, "(".ToCharArray());
				if(ast.l.leaf){
					tf.stringx = AppendString(tf.stringx, ast.l.value);
				}else{
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					CreateStringFromNumberWithCheck(tl, 10d, numberString);
					tf.stringx = AppendString(tf.stringx, numberString.stringx);
				}
				tf.stringx = AppendString(tf.stringx, ")".ToCharArray());
				tf.stringx = AppendString(tf.stringx, ";\n".ToCharArray());
			}else if(arraysStringsEqual(ast.value, "%".ToCharArray())){
				t.numberValue = t.numberValue + 1d;
				tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
				CreateStringFromNumberWithCheck(t.numberValue, 10d, numberString);
				tf.stringx = AppendString(tf.stringx, numberString.stringx);
				tf.stringx = AppendString(tf.stringx, " = Percentage(".ToCharArray());
				tf.stringx = AppendString(tf.stringx, ast.l.value);
				tf.stringx = AppendString(tf.stringx, ")".ToCharArray());
				tf.stringx = AppendString(tf.stringx, ";\n".ToCharArray());
			}else{
				tf.stringx = AppendString(tf.stringx, "<failed>".ToCharArray());
			}
		}
	}


	public static char [] ArithmeticFormulaToTForm(char [] f){
		char [] af;
		StringArrayReference tokens;
		bool success;
		StringReference errorMessage, tf;
		ASTNode ast;
		NumberReference pos, t;

		tokens = new StringArrayReference();
		errorMessage = new StringReference();
		success = TokenizeArithmeticFormula(f, tokens, errorMessage);

		if(success){
			/* Parse*/
			ast = new ASTNode();
			pos = CreateNumberReference(0d);
			success = ParseArithmeticTokens(tokens, pos, ast, errorMessage);

			if(success){
				tf = CreateStringReference("".ToCharArray());
				t = CreateNumberReference(0d);
				ArithmeticASTToTForm(ast, tf, t);
				af = tf.stringx;
			}else{
				af = errorMessage.stringx;
			}
		}else{
			af = errorMessage.stringx;
		}

		return af;
	}


	public static char [] ASTToBitwiseFormula(ASTNode ast){
		char [] f, p1, p2;

		if(ast.leaf){
			f = ast.value;
		}else if(arraysStringsEqual(ast.value, "~".ToCharArray())){
			f = new char [0];
			p1 = ASTToBitwiseFormula(ast.l);
			f = AppendString(f, "~".ToCharArray());
			f = AppendString(f, p1);
		}else if(arraysStringsEqual(ast.value, "()".ToCharArray())){
			f = new char [0];
			p1 = ASTToBitwiseFormula(ast.l);
			f = AppendString(f, "(".ToCharArray());
			f = AppendString(f, p1);
			f = AppendString(f, ")".ToCharArray());
		}else if(arraysStringsEqual(ast.value, "&".ToCharArray()) || arraysStringsEqual(ast.value, "|".ToCharArray()) || arraysStringsEqual(ast.value, "^".ToCharArray()) || arraysStringsEqual(ast.value, "<<".ToCharArray()) || arraysStringsEqual(ast.value, ">>".ToCharArray())){
			f = new char [0];
			p1 = ASTToBitwiseFormula(ast.l);
			p2 = ASTToBitwiseFormula(ast.r);
			f = AppendString(f, p1);
			f = AppendString(f, " ".ToCharArray());
			f = AppendString(f, ast.value);
			f = AppendString(f, " ".ToCharArray());
			f = AppendString(f, p2);
		}else{
			f = "<failed>".ToCharArray();
		}

		return f;
	}


	public static void BitwiseASTToTForm(ASTNode ast, StringReference tf, NumberReference t){
		double tl, tr;
		StringReference numberString;
		numberString = new StringReference();

		tl = 0d;
		tr = 0d;

		if(!ast.leaf){
			if(arraysStringsEqual(ast.value, "&".ToCharArray()) || arraysStringsEqual(ast.value, "|".ToCharArray()) || arraysStringsEqual(ast.value, "^".ToCharArray()) || arraysStringsEqual(ast.value, "<<".ToCharArray()) || arraysStringsEqual(ast.value, ">>".ToCharArray())){
				if(!ast.l.leaf){
					BitwiseASTToTForm(ast.l, tf, t);
					tl = t.numberValue;
				}

				if(!ast.r.leaf){
					BitwiseASTToTForm(ast.r, tf, t);
					tr = t.numberValue;
				}

				t.numberValue = t.numberValue + 1d;
				tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
				CreateStringFromNumberWithCheck(t.numberValue, 10d, numberString);
				tf.stringx = AppendString(tf.stringx, numberString.stringx);
				tf.stringx = AppendString(tf.stringx, " = ".ToCharArray());
				if(ast.l.leaf){
					tf.stringx = AppendString(tf.stringx, ast.l.value);
				}else{
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					CreateStringFromNumberWithCheck(tl, 10d, numberString);
					tf.stringx = AppendString(tf.stringx, numberString.stringx);
				}
				tf.stringx = AppendString(tf.stringx, " ".ToCharArray());
				tf.stringx = AppendString(tf.stringx, ast.value);
				tf.stringx = AppendString(tf.stringx, " ".ToCharArray());
				if(ast.r.leaf){
					tf.stringx = AppendString(tf.stringx, ast.r.value);
				}else{
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					CreateStringFromNumberWithCheck(tr, 10d, numberString);
					tf.stringx = AppendString(tf.stringx, numberString.stringx);
				}
				tf.stringx = AppendString(tf.stringx, ";\n".ToCharArray());
			}else if(arraysStringsEqual(ast.value, "()".ToCharArray())){
				BitwiseASTToTForm(ast.l, tf, t);
				tl = t.numberValue;

				t.numberValue = t.numberValue + 1d;
				tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
				CreateStringFromNumberWithCheck(t.numberValue, 10d, numberString);
				tf.stringx = AppendString(tf.stringx, numberString.stringx);
				tf.stringx = AppendString(tf.stringx, " = ".ToCharArray());
				tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
				CreateStringFromNumberWithCheck(tl, 10d, numberString);
				tf.stringx = AppendString(tf.stringx, numberString.stringx);
				tf.stringx = AppendString(tf.stringx, ";\n".ToCharArray());
			}else if(arraysStringsEqual(ast.value, "!".ToCharArray())){
				if(!ast.l.leaf){
					BitwiseASTToTForm(ast.l, tf, t);
					tl = t.numberValue;
				}

				t.numberValue = t.numberValue + 1d;
				tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
				CreateStringFromNumberWithCheck(t.numberValue, 10d, numberString);
				tf.stringx = AppendString(tf.stringx, numberString.stringx);
				tf.stringx = AppendString(tf.stringx, " = ".ToCharArray());
				tf.stringx = AppendString(tf.stringx, "!".ToCharArray());
				if(ast.l.leaf){
					tf.stringx = AppendString(tf.stringx, ast.l.value);
				}else{
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					CreateStringFromNumberWithCheck(tl, 10d, numberString);
					tf.stringx = AppendString(tf.stringx, numberString.stringx);
				}
				tf.stringx = AppendString(tf.stringx, ";\n".ToCharArray());
			}else{
				tf.stringx = AppendString(tf.stringx, "<failed>".ToCharArray());
			}
		}
	}


	public static char [] BitwiseFormulaToTForm(char [] f){
		char [] af;
		StringArrayReference tokens;
		bool success;
		StringReference errorMessage, tf;
		ASTNode ast;
		NumberReference pos, t;

		tokens = new StringArrayReference();
		errorMessage = new StringReference();
		success = TokenizeBitwiseFormula(f, tokens, errorMessage);

		if(success){
			/* Parse*/
			ast = new ASTNode();
			pos = CreateNumberReference(0d);
			success = ParseBitwiseTokens(tokens, pos, ast, errorMessage);

			if(success){
				tf = CreateStringReference("".ToCharArray());
				t = CreateNumberReference(0d);
				BitwiseASTToTForm(ast, tf, t);
				af = tf.stringx;
			}else{
				af = errorMessage.stringx;
			}
		}else{
			af = errorMessage.stringx;
		}

		return af;
	}


	public static void BooleanASTToTFormFunctions(ASTNode ast, StringReference tf, char [] prefix, char [] postfix, char [] tprefix, bool parenthesis, bool semicolon, bool wrappedNumber, char [] target, bool endsWithIf){
		BooleanArrayReference t;
		NumberReference assignedT;

		t = CreateBooleanArrayReferenceLengthValue(0d, false);
		assignedT = CreateNumberReference(0d);
		tf.stringx = new char [0];

		BooleanASTToTFormFunctionsInner(ast, tf, t, assignedT, prefix, postfix, tprefix, parenthesis, semicolon, wrappedNumber, target, 0d, endsWithIf);
	}


	public static void BooleanASTToTFormFunctionsInner(ASTNode ast, StringReference tf, BooleanArrayReference ts, NumberReference assignedT, char [] prefix, char [] postfix, char [] tprefix, bool parenthesis, bool semicolon, bool wrappedNumber, char [] target, double level, bool endsWithIf){
		double tl, tr, t;
		char [] functionName, value, numberString;

		tl = 0d;
		tr = 0d;
		t = 0d;

		if(!ast.leaf){
			if(arraysStringsEqual(ast.value, "&".ToCharArray()) || arraysStringsEqual(ast.value, "|".ToCharArray()) || arraysStringsEqual(ast.value, "^".ToCharArray()) || arraysStringsEqual(ast.value, "=".ToCharArray()) || arraysStringsEqual(ast.value, "!=".ToCharArray()) || arraysStringsEqual(ast.value, "<".ToCharArray()) || arraysStringsEqual(ast.value, "<=".ToCharArray()) || arraysStringsEqual(ast.value, ">".ToCharArray()) || arraysStringsEqual(ast.value, ">=".ToCharArray()) || arraysStringsEqual(ast.value, "->".ToCharArray()) || arraysStringsEqual(ast.value, "<->".ToCharArray())){
				if(!ast.l.leaf){
					BooleanASTToTFormFunctionsInner(ast.l, tf, ts, assignedT, prefix, postfix, tprefix, parenthesis, semicolon, wrappedNumber, target, level + 1d, endsWithIf);
					tl = assignedT.numberValue;
				}

				if(!ast.r.leaf){
					BooleanASTToTFormFunctionsInner(ast.r, tf, ts, assignedT, prefix, postfix, tprefix, parenthesis, semicolon, wrappedNumber, target, level + 1d, endsWithIf);
					tr = assignedT.numberValue;
				}

				if(!ast.l.leaf){
					FreeTVariable(ts, tl);
				}

				if(!ast.r.leaf){
					FreeTVariable(ts, tr);
				}

				functionName = BooleanBinarySymbolToFunctionName(ast.value);

				tf.stringx = AppendString(tf.stringx, prefix);
				tf.stringx = AppendString(tf.stringx, functionName);
				tf.stringx = AppendString(tf.stringx, postfix);
				if(parenthesis){
					tf.stringx = AppendString(tf.stringx, "(".ToCharArray());
				}else{
					tf.stringx = AppendString(tf.stringx, " ".ToCharArray());
				}

				if(level > 0d || target.Length == 0d){
					t = AllocateTVariable(ts);
					assignedT.numberValue = t;
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					tf.stringx = AppendString(tf.stringx, tprefix);
					numberString = CreateStringDecimalFromNumber(t);
					tf.stringx = AppendString(tf.stringx, numberString);
				}else{
					tf.stringx = AppendString(tf.stringx, target);
				}

				tf.stringx = AppendString(tf.stringx, ", ".ToCharArray());

				if(ast.l.leaf){
					if(charIsNumber(ast.l.value[0])){
						value = new char [0];
						value = AppendString(value, prefix);
						if(wrappedNumber){
							value = AppendString(value, "Number".ToCharArray());
							value = AppendString(value, postfix);
							value = AppendString(value, "(".ToCharArray());
						}
						value = AppendString(value, ast.l.value);
						if(wrappedNumber){
							value = AppendString(value, ")".ToCharArray());
						}
					}else{
						value = ast.l.value;
					}
					tf.stringx = AppendString(tf.stringx, value);
				}else{
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					tf.stringx = AppendString(tf.stringx, tprefix);
					numberString = CreateStringDecimalFromNumber(tl);
					tf.stringx = AppendString(tf.stringx, numberString);
				}
				tf.stringx = AppendString(tf.stringx, ", ".ToCharArray());
				if(ast.r.leaf){
					if(charIsNumber(ast.r.value[0])){
						value = new char [0];
						value = AppendString(value, prefix);
						if(wrappedNumber){
							value = AppendString(value, "Number".ToCharArray());
							value = AppendString(value, postfix);
							value = AppendString(value, "(".ToCharArray());
						}
						value = AppendString(value, ast.r.value);
						if(wrappedNumber){
							value = AppendString(value, ")".ToCharArray());
						}
					}else{
						value = ast.r.value;
					}
					tf.stringx = AppendString(tf.stringx, value);
				}else{
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					tf.stringx = AppendString(tf.stringx, tprefix);
					numberString = CreateStringDecimalFromNumber(tr);
					tf.stringx = AppendString(tf.stringx, numberString);
				}
				if(parenthesis){
					tf.stringx = AppendString(tf.stringx, ")".ToCharArray());
				}else{
					tf.stringx = AppendString(tf.stringx, " ".ToCharArray());
				}
				if(semicolon){
					tf.stringx = AppendString(tf.stringx, ";".ToCharArray());
				}
				tf.stringx = AppendString(tf.stringx, "\n".ToCharArray());
			}else if(arraysStringsEqual(ast.value, "()".ToCharArray())){
				BooleanASTToTFormFunctionsInner(ast.l, tf, ts, assignedT, prefix, postfix, tprefix, parenthesis, semicolon, wrappedNumber, target, level + 1d, endsWithIf);
			}else if(arraysStringsEqual(ast.value, "!".ToCharArray())){
				if(!ast.l.leaf){
					BooleanASTToTFormFunctionsInner(ast.l, tf, ts, assignedT, prefix, postfix, tprefix, parenthesis, semicolon, wrappedNumber, target, level + 1d, endsWithIf);
					tl = assignedT.numberValue;
					FreeTVariable(ts, tl);
				}

				functionName = BooleanBinarySymbolToFunctionName(ast.value);

				tf.stringx = AppendString(tf.stringx, prefix);
				tf.stringx = AppendString(tf.stringx, functionName);
				tf.stringx = AppendString(tf.stringx, postfix);
				if(parenthesis){
					tf.stringx = AppendString(tf.stringx, "(".ToCharArray());
				}else{
					tf.stringx = AppendString(tf.stringx, " ".ToCharArray());
				}

				if(level > 0d || target.Length == 0d){
					t = AllocateTVariable(ts);
					assignedT.numberValue = t;
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					tf.stringx = AppendString(tf.stringx, tprefix);
					numberString = CreateStringDecimalFromNumber(t);
					tf.stringx = AppendString(tf.stringx, numberString);
				}else{
					tf.stringx = AppendString(tf.stringx, target);
				}

				tf.stringx = AppendString(tf.stringx, ", ".ToCharArray());

				if(ast.l.leaf){
					tf.stringx = AppendString(tf.stringx, ast.l.value);
				}else{
					tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
					tf.stringx = AppendString(tf.stringx, tprefix);
					numberString = CreateStringDecimalFromNumber(tl);
					tf.stringx = AppendString(tf.stringx, numberString);
				}
				if(parenthesis){
					tf.stringx = AppendString(tf.stringx, ")".ToCharArray());
				}else{
					tf.stringx = AppendString(tf.stringx, " ".ToCharArray());
				}
				if(semicolon){
					tf.stringx = AppendString(tf.stringx, ";".ToCharArray());
				}
				tf.stringx = AppendString(tf.stringx, "\n".ToCharArray());
			}else{
				tf.stringx = AppendString(tf.stringx, "<failed>".ToCharArray());
			}
		}else{
			tf.stringx = AppendString(tf.stringx, "<failed>".ToCharArray());
		}

		if(level == 0d && endsWithIf){
			tf.stringx = AppendString(tf.stringx, "If ".ToCharArray());
			if(target.Length == 0d){
				tf.stringx = AppendString(tf.stringx, "t".ToCharArray());
				tf.stringx = AppendString(tf.stringx, tprefix);
				numberString = CreateStringDecimalFromNumber(t);
				tf.stringx = AppendString(tf.stringx, numberString);
			}else{
				tf.stringx = AppendString(tf.stringx, target);
			}
		}
	}


	public static char [] BooleanBinarySymbolToFunctionName(char [] value){
		char [] f;

		f = "Unknown".ToCharArray();

		if(arraysStringsEqual(value, "!".ToCharArray())){
			f = "Not".ToCharArray();
		}
		if(arraysStringsEqual(value, "&".ToCharArray())){
			f = "And".ToCharArray();
		}
		if(arraysStringsEqual(value, "|".ToCharArray())){
			f = "Or".ToCharArray();
		}
		if(arraysStringsEqual(value, "^".ToCharArray())){
			f = "Xor".ToCharArray();
		}
		if(arraysStringsEqual(value, "=".ToCharArray())){
			f = "Equal".ToCharArray();
		}
		if(arraysStringsEqual(value, "!=".ToCharArray())){
			f = "Unequal".ToCharArray();
		}
		if(arraysStringsEqual(value, "<".ToCharArray())){
			f = "LessThan".ToCharArray();
		}
		if(arraysStringsEqual(value, "<=".ToCharArray())){
			f = "LessThanOrEqual".ToCharArray();
		}
		if(arraysStringsEqual(value, ">".ToCharArray())){
			f = "MoreThan".ToCharArray();
		}
		if(arraysStringsEqual(value, ">=".ToCharArray())){
			f = "MoreThanOrEqual".ToCharArray();
		}
		if(arraysStringsEqual(value, "->".ToCharArray())){
			f = "Implies".ToCharArray();
		}
		if(arraysStringsEqual(value, "<->".ToCharArray())){
			f = "Iff".ToCharArray();
		}

		return f;
	}


	public static bool BooleanFormulaToTFormFunctions(char [] f, char [] prefix, char [] postfix, char [] tprefix, bool parenthesis, bool semicolon, bool wrappedNumber, char [] target, bool endsWithIf, StringReference tf, StringReference message){
		StringArrayReference tokens;
		bool success;
		ASTNode ast;
		NumberReference pos;

		tokens = new StringArrayReference();
		success = TokenizeBooleanFormula(f, tokens, message);

		if(success){
			/* Parse*/
			ast = new ASTNode();
			pos = CreateNumberReference(0d);
			success = ParseBooleanTokens(tokens, pos, ast, message);

			if(success){
				BooleanASTToTFormFunctions(ast, tf, prefix, postfix, tprefix, parenthesis, semicolon, wrappedNumber, target, endsWithIf);
			}
		}

		return success;
	}


	public static BooleanReference CreateBooleanReference(bool value){
		BooleanReference refx;

		refx = new BooleanReference();
		refx.booleanValue = value;

		return refx;
	}


	public static BooleanArrayReference CreateBooleanArrayReference(bool [] value){
		BooleanArrayReference refx;

		refx = new BooleanArrayReference();
		refx.booleanArray = value;

		return refx;
	}


	public static BooleanArrayReference CreateBooleanArrayReferenceLengthValue(double length, bool value){
		BooleanArrayReference refx;
		double i;

		refx = new BooleanArrayReference();
		refx.booleanArray = new bool [(int)(length)];

		for(i = 0d; i < length; i = i + 1d){
			refx.booleanArray[(int)(i)] = value;
		}

		return refx;
	}


	public static void FreeBooleanArrayReference(BooleanArrayReference booleanArrayReference){
		delete(booleanArrayReference.booleanArray);
		delete(booleanArrayReference);
	}


	public static CharacterReference CreateCharacterReference(char value){
		CharacterReference refx;

		refx = new CharacterReference();
		refx.characterValue = value;

		return refx;
	}


	public static NumberReference CreateNumberReference(double value){
		NumberReference refx;

		refx = new NumberReference();
		refx.numberValue = value;

		return refx;
	}


	public static NumberArrayReference CreateNumberArrayReference(double [] value){
		NumberArrayReference refx;

		refx = new NumberArrayReference();
		refx.numberArray = value;

		return refx;
	}


	public static NumberArrayReference CreateNumberArrayReferenceLengthValue(double length, double value){
		NumberArrayReference refx;
		double i;

		refx = new NumberArrayReference();
		refx.numberArray = new double [(int)(length)];

		for(i = 0d; i < length; i = i + 1d){
			refx.numberArray[(int)(i)] = value;
		}

		return refx;
	}


	public static void FreeNumberArrayReference(NumberArrayReference numberArrayReference){
		delete(numberArrayReference.numberArray);
		delete(numberArrayReference);
	}


	public static StringReference CreateStringReference(char [] value){
		StringReference refx;

		refx = new StringReference();
		refx.stringx = value;

		return refx;
	}


	public static StringReference CreateStringReferenceLengthValue(double length, char value){
		StringReference refx;
		double i;

		refx = new StringReference();
		refx.stringx = new char [(int)(length)];

		for(i = 0d; i < length; i = i + 1d){
			refx.stringx[(int)(i)] = value;
		}

		return refx;
	}


	public static void FreeStringReference(StringReference stringReference){
		delete(stringReference.stringx);
		delete(stringReference);
	}


	public static StringArrayReference CreateStringArrayReference(StringReference [] strings){
		StringArrayReference refx;

		refx = new StringArrayReference();
		refx.stringArray = strings;

		return refx;
	}


	public static StringArrayReference CreateStringArrayReferenceLengthValue(double length, char [] value){
		StringArrayReference refx;
		double i;

		refx = new StringArrayReference();
		refx.stringArray = new StringReference [(int)(length)];

		for(i = 0d; i < length; i = i + 1d){
			refx.stringArray[(int)(i)] = CreateStringReference(value);
		}

		return refx;
	}


	public static void FreeStringArrayReference(StringArrayReference stringArrayReference){
		double i;

		for(i = 0d; i < stringArrayReference.stringArray.Length; i = i + 1d){
			delete(stringArrayReference.stringArray[(int)(i)]);
		}
		delete(stringArrayReference.stringArray);
		delete(stringArrayReference);
	}


	public static char [] CreateStringScientificNotationDecimalFromNumber(double n){
		StringReference mantissaReference, exponentReference;
		double e;
		bool isPositive;
		char [] result;

		mantissaReference = new StringReference();
		exponentReference = new StringReference();
		result = new char [0];

		if(n < 0d){
			isPositive = false;
			n = -n;
		}else{
			isPositive = true;
		}

		if(n == 0d){
			e = 0d;
		}else{
			e = GetFirstDecimalDigitPosition(n);

			if(e < 0d){
				n = n*Pow(10d, Abs(e));
			}else{
				n = n/Pow(10d, e);
			}
		}

		mantissaReference.stringx = CreateStringDecimalFromNumber(n);
		exponentReference.stringx = CreateStringDecimalFromNumber(e);

		if(!isPositive){
			result = AppendString(result, "-".ToCharArray());
		}

		result = AppendString(result, mantissaReference.stringx);
		result = AppendString(result, "e".ToCharArray());
		result = AppendString(result, exponentReference.stringx);

		return result;
	}


	public static char [] CreateStringDecimalFromNumber(double number){
		DynamicArrayCharacters stringx;
		double maximumDigits, i, d, digitPosition, trailingZeros;
		bool hasPrintedPoint, isPositive, done;
		CharacterReference characterReference;
		char c;
		char [] str;

		stringx = CreateDynamicArrayCharacters();
		isPositive = true;

		if(number < 0d){
			isPositive = false;
			number = -number;
		}

		if(number == 0d){
			DynamicArrayAddCharacter(stringx, '0');
		}else{
			characterReference = new CharacterReference();

			maximumDigits = GetMaximumDigitsForDecimal();

			digitPosition = GetFirstDecimalDigitPosition(number);

			hasPrintedPoint = false;

			if(!isPositive){
				DynamicArrayAddCharacter(stringx, '-');
			}

			/* Print leading zeros.*/
			if(digitPosition < 0d){
				DynamicArrayAddCharacter(stringx, '0');
				DynamicArrayAddCharacter(stringx, '.');
				hasPrintedPoint = true;
				for(i = 0d; i < -digitPosition - 1d; i = i + 1d){
					DynamicArrayAddCharacter(stringx, '0');
				}
			}

			/* Count trailing zeros*/
			trailingZeros = 0d;
			done = false;
			for(i = 0d; i < maximumDigits && !done; i = i + 1d){
				d = GetDecimalDigitWithFirstDigitPosition(number, digitPosition, maximumDigits - i - 1d);
				if(d == 0d){
					trailingZeros = trailingZeros + 1d;
				}else{
					done = true;
				}
			}

			/* Print number.*/
			for(i = 0d; i < maximumDigits; i = i + 1d){
				d = GetDecimalDigitWithFirstDigitPosition(number, digitPosition, i);

				if(!hasPrintedPoint && digitPosition - i + 1d == 0d){
					if(maximumDigits - i > trailingZeros){
						DynamicArrayAddCharacter(stringx, '.');
					}
					hasPrintedPoint = true;
				}

				if(maximumDigits - i <= trailingZeros && hasPrintedPoint){
				}else{
					GetDecimalDigitCharacterFromNumberWithCheck(d, characterReference);
					c = characterReference.characterValue;
					DynamicArrayAddCharacter(stringx, c);
				}
			}

			/* Print trailing zeros.*/
			for(i = 0d; i < digitPosition - maximumDigits + 1d; i = i + 1d){
				DynamicArrayAddCharacter(stringx, '0');
			}
		}

		/* Done*/
		str = DynamicArrayCharactersToArray(stringx);
		FreeDynamicArrayCharacters(stringx);
		return str;
	}


	public static bool CreateStringFromNumberWithCheck(double number, double basex, StringReference stringRef){
		DynamicArrayCharacters stringx;
		double maximumDigits, i, d, digitPosition, trailingZeros;
		bool success, hasPrintedPoint, isPositive, done;
		CharacterReference characterReference;
		char c;

		stringx = CreateDynamicArrayCharacters();
		isPositive = true;

		if(number < 0d){
			isPositive = false;
			number = -number;
		}

		if(number == 0d){
			DynamicArrayAddCharacter(stringx, '0');
			success = true;
		}else{
			characterReference = new CharacterReference();

			if(m18IsInteger(basex)){
				success = true;

				maximumDigits = GetMaximumDigitsForBase(basex);

				digitPosition = GetFirstDigitPosition(number, basex);

				hasPrintedPoint = false;

				if(!isPositive){
					DynamicArrayAddCharacter(stringx, '-');
				}

				/* Print leading zeros.*/
				if(digitPosition < 0d){
					DynamicArrayAddCharacter(stringx, '0');
					DynamicArrayAddCharacter(stringx, '.');
					hasPrintedPoint = true;
					for(i = 0d; i < -digitPosition - 1d; i = i + 1d){
						DynamicArrayAddCharacter(stringx, '0');
					}
				}

				/* Count trailing zeros*/
				trailingZeros = 0d;
				done = false;
				for(i = 0d; i < maximumDigits && !done; i = i + 1d){
					d = GetDigit(number, basex, maximumDigits - i - 1d);
					if(d == 0d){
						trailingZeros = trailingZeros + 1d;
					}else{
						done = true;
					}
				}

				/* Print number.*/
				for(i = 0d; i < maximumDigits && success; i = i + 1d){
					d = GetDigit(number, basex, i);

					if(d >= basex){
						d = basex - 1d;
					}

					if(!hasPrintedPoint && digitPosition - i + 1d == 0d){
						if(maximumDigits - i > trailingZeros){
							DynamicArrayAddCharacter(stringx, '.');
						}
						hasPrintedPoint = true;
					}

					if(maximumDigits - i <= trailingZeros && hasPrintedPoint){
					}else{
						success = GetSingleDigitCharacterFromNumberWithCheck(d, basex, characterReference);
						if(success){
							c = characterReference.characterValue;
							DynamicArrayAddCharacter(stringx, c);
						}
					}
				}

				if(success){
					/* Print trailing zeros.*/
					for(i = 0d; i < digitPosition - maximumDigits + 1d; i = i + 1d){
						DynamicArrayAddCharacter(stringx, '0');
					}
				}
			}else{
				success = false;
			}
		}

		if(success){
			stringRef.stringx = DynamicArrayCharactersToArray(stringx);
			FreeDynamicArrayCharacters(stringx);
		}

		/* Done*/
		return success;
	}


	public static double GetMaximumDigitsForBase(double basex){
		double t;

		t = Pow(10d, 15d);
		return Floor(Log10(t)/Log10(basex));
	}


	public static double GetMaximumDigitsForDecimal(){
		return 15d;
	}


	public static double GetFirstDecimalDigitPosition(double n){
		double power, m, i;
		bool multiply, done;

		n = Abs(n);

		if(n != 0d){
			if(Floor(n) < Pow(10d, 15d)){
				multiply = true;
			}else{
				multiply = false;
			}

			done = false;
			m = 0d;
			for(i = 0d; !done; i = i + 1d){
				if(multiply){
					m = n*Pow(10d, i);
					if(Floor(m) >= Pow(10d, 14d)){
						done = true;
					}
				}else{
					m = n/Pow(10d, i);
					if(Floor(m) < Pow(10d, 15d)){
						done = true;
					}
				}
			}

			if(multiply){
				power = 15d - i;
			}else{
				power = 15d + i - 2d;
			}

			if(m18Round(m) >= Pow(10d, 15d)){
				power = power + 1d;
			}
		}else{
			power = 1d;
		}

		return power;
	}


	public static double GetFirstDigitPosition(double n, double basex){
		double power, m, i, maximumDigits;
		bool multiply, done;

		maximumDigits = GetMaximumDigitsForBase(basex);
		n = Abs(n);

		if(n != 0d){
			if(Floor(n) < Pow(basex, maximumDigits)){
				multiply = true;
			}else{
				multiply = false;
			}

			done = false;
			m = 0d;
			for(i = 0d; !done; i = i + 1d){
				if(multiply){
					m = n*Pow(basex, i);
					if(Floor(m) >= Pow(basex, maximumDigits - 1d)){
						done = true;
					}
				}else{
					m = n/Pow(basex, i);
					if(Floor(m) < Pow(basex, maximumDigits)){
						done = true;
					}
				}
			}

			if(multiply){
				power = maximumDigits - i;
			}else{
				power = maximumDigits + i - 2d;
			}

			if(m18Round(m) >= Pow(basex, maximumDigits)){
				power = power + 1d;
			}
		}else{
			power = 1d;
		}

		return power;
	}


	public static bool GetSingleDigitCharacterFromNumberWithCheck(double c, double basex, CharacterReference characterReference){
		char [] numberTable;
		bool success;

		numberTable = GetDigitCharacterTable();

		if(c < basex || c < numberTable.Length){
			success = true;
			characterReference.characterValue = numberTable[(int)(c)];
		}else{
			success = false;
		}

		return success;
	}


	public static bool GetDecimalDigitCharacterFromNumberWithCheck(double c, CharacterReference characterRef){
		char [] numberTable;
		bool success;

		numberTable = "0123456789".ToCharArray();

		if(c >= 0d && c < 10d){
			success = true;
			characterRef.characterValue = numberTable[(int)(c)];
		}else{
			success = false;
		}

		return success;
	}


	public static char [] GetDigitCharacterTable(){
		char [] numberTable;

		numberTable = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ".ToCharArray();

		return numberTable;
	}


	public static double GetDecimalDigit(double n, double index){
		double digitPosition;

		digitPosition = GetFirstDecimalDigitPosition(n);

		return GetDecimalDigitWithFirstDigitPosition(n, digitPosition, index);
	}


	public static double GetDecimalDigitWithFirstDigitPosition(double n, double digitPosition, double index){
		double d, e, m, i;

		n = Abs(n);

		e = 15d - digitPosition - 1d;
		if(e < 0d){
			n = (double)Round(n/Pow(10d, Abs(e)));
		}else{
			n = (double)Round(n*Pow(10d, e));
		}

		m = n;
		d = 0d;
		for(i = 0d; i < 15d - index; i = i + 1d){
			d = (double)Round(m%10d);
			m = m - d;
			m = (double)Round(m/10d);
		}

		return d;
	}


	public static double GetDigit(double n, double basex, double index){
		double d, digitPosition, e, m, maximumDigits, i;

		n = Abs(n);
		maximumDigits = GetMaximumDigitsForBase(basex);
		digitPosition = GetFirstDigitPosition(n, basex);

		e = maximumDigits - digitPosition - 1d;
		if(e < 0d){
			n = (double)Round(n/Pow(basex, Abs(e)));
		}else{
			n = (double)Round(n*Pow(basex, e));
		}

		m = n;
		d = 0d;
		for(i = 0d; i < maximumDigits - index; i = i + 1d){
			d = (double)Round(m%basex);
			m = m - d;
			m = (double)Round(m/basex);
		}

		return d;
	}


	public static bool CreateNumberFromDecimalStringWithCheck(char [] stringx, NumberReference decimalReference, StringReference message){
		return CreateNumberFromStringWithCheck(stringx, 10d, decimalReference, message);
	}


	public static double CreateNumberFromDecimalString(char [] stringx){
		NumberReference doubleReference;
		StringReference stringReference;
		double number;

		doubleReference = CreateNumberReference(0d);
		stringReference = CreateStringReference("".ToCharArray());
		CreateNumberFromStringWithCheck(stringx, 10d, doubleReference, stringReference);
		number = doubleReference.numberValue;

		delete(doubleReference);
		delete(stringReference);

		return number;
	}


	public static bool CreateNumberFromStringWithCheck(char [] stringx, double basex, NumberReference numberReference, StringReference message){
		bool success;
		BooleanReference numberIsPositive, exponentIsPositive;
		NumberArrayReference beforePoint, afterPoint, exponent;

		numberIsPositive = CreateBooleanReference(true);
		exponentIsPositive = CreateBooleanReference(true);
		beforePoint = new NumberArrayReference();
		afterPoint = new NumberArrayReference();
		exponent = new NumberArrayReference();

		if(basex >= 2d && basex <= 36d){
			success = ExtractPartsFromNumberString(stringx, basex, numberIsPositive, beforePoint, afterPoint, exponentIsPositive, exponent, message);

			if(success){
				numberReference.numberValue = CreateNumberFromParts(basex, numberIsPositive.booleanValue, beforePoint.numberArray, afterPoint.numberArray, exponentIsPositive.booleanValue, exponent.numberArray);
			}
		}else{
			success = false;
			message.stringx = "Base must be from 2 to 36.".ToCharArray();
		}

		return success;
	}


	public static double CreateNumberFromParts(double basex, bool numberIsPositive, double [] beforePoint, double [] afterPoint, bool exponentIsPositive, double [] exponent){
		double n, i, p, e;

		n = 0d;

		for(i = 0d; i < beforePoint.Length; i = i + 1d){
			p = beforePoint[(int)(beforePoint.Length - i - 1d)];

			n = n + p*Pow(basex, i);
		}

		for(i = 0d; i < afterPoint.Length; i = i + 1d){
			p = afterPoint[(int)(i)];

			n = n + p/Pow(basex, i + 1d);
		}

		if(exponent.Length > 0d){
			e = 0d;
			for(i = 0d; i < exponent.Length; i = i + 1d){
				p = exponent[(int)(exponent.Length - i - 1d)];

				e = e + p*Pow(basex, i);
			}

			if(!exponentIsPositive){
				e = -e;
			}

			n = n*Pow(basex, e);
		}

		if(!numberIsPositive){
			n = -n;
		}

		return n;
	}


	public static bool ExtractPartsFromNumberString(char [] n, double basex, BooleanReference numberIsPositive, NumberArrayReference beforePoint, NumberArrayReference afterPoint, BooleanReference exponentIsPositive, NumberArrayReference exponent, StringReference errorMessages){
		double i, j, count;
		bool success, done, complete;

		i = 0d;
		complete = false;

		if(i < n.Length){
			if(n[(int)(i)] == '-'){
				numberIsPositive.booleanValue = false;
				i = i + 1d;
			}else if(n[(int)(i)] == '+'){
				numberIsPositive.booleanValue = true;
				i = i + 1d;
			}

			success = true;
		}else{
			success = false;
			errorMessages.stringx = "Number cannot have length zero.".ToCharArray();
		}

		if(success){
			done = false;
			count = 0d;
			for(; i + count < n.Length && !done; ){
				if(CharacterIsNumberCharacterInBase(n[(int)(i + count)], basex)){
					count = count + 1d;
				}else{
					done = true;
				}
			}

			if(count >= 1d){
				beforePoint.numberArray = new double [(int)(count)];

				for(j = 0d; j < count; j = j + 1d){
					beforePoint.numberArray[(int)(j)] = GetNumberFromNumberCharacterForBase(n[(int)(i + j)], basex);
				}

				i = i + count;

				if(i < n.Length){
					success = true;
				}else{
					afterPoint.numberArray = new double [0];
					exponent.numberArray = new double [0];
					success = true;
					complete = true;
				}
			}else{
				success = false;
				errorMessages.stringx = "Number must have at least one number after the optional sign.".ToCharArray();
			}
		}

		if(success && !complete){
			if(n[(int)(i)] == '.'){
				i = i + 1d;

				if(i < n.Length){
					done = false;
					count = 0d;
					for(; i + count < n.Length && !done; ){
						if(CharacterIsNumberCharacterInBase(n[(int)(i + count)], basex)){
							count = count + 1d;
						}else{
							done = true;
						}
					}

					if(count >= 1d){
						afterPoint.numberArray = new double [(int)(count)];

						for(j = 0d; j < count; j = j + 1d){
							afterPoint.numberArray[(int)(j)] = GetNumberFromNumberCharacterForBase(n[(int)(i + j)], basex);
						}

						i = i + count;

						if(i < n.Length){
							success = true;
						}else{
							exponent.numberArray = new double [0];
							success = true;
							complete = true;
						}
					}else{
						success = false;
						errorMessages.stringx = "There must be at least one digit after the decimal point.".ToCharArray();
					}
				}else{
					success = false;
					errorMessages.stringx = "There must be at least one digit after the decimal point.".ToCharArray();
				}
			}else if(basex <= 14d && (n[(int)(i)] == 'e' || n[(int)(i)] == 'E')){
				if(i < n.Length){
					success = true;
					afterPoint.numberArray = new double [0];
				}else{
					success = false;
					errorMessages.stringx = "There must be at least one digit after the exponent.".ToCharArray();
				}
			}else{
				success = false;
				errorMessages.stringx = "Expected decimal point or exponent symbol.".ToCharArray();
			}
		}

		if(success && !complete){
			if(basex <= 14d && (n[(int)(i)] == 'e' || n[(int)(i)] == 'E')){
				i = i + 1d;

				if(i < n.Length){
					if(n[(int)(i)] == '-'){
						exponentIsPositive.booleanValue = false;
						i = i + 1d;
					}else if(n[(int)(i)] == '+'){
						exponentIsPositive.booleanValue = true;
						i = i + 1d;
					}

					if(i < n.Length){
						done = false;
						count = 0d;
						for(; i + count < n.Length && !done; ){
							if(CharacterIsNumberCharacterInBase(n[(int)(i + count)], basex)){
								count = count + 1d;
							}else{
								done = true;
							}
						}

						if(count >= 1d){
							exponent.numberArray = new double [(int)(count)];

							for(j = 0d; j < count; j = j + 1d){
								exponent.numberArray[(int)(j)] = GetNumberFromNumberCharacterForBase(n[(int)(i + j)], basex);
							}

							i = i + count;

							if(i == n.Length){
								success = true;
							}else{
								success = false;
								errorMessages.stringx = "There cannot be any characters past the exponent of the number.".ToCharArray();
							}
						}else{
							success = false;
							errorMessages.stringx = "There must be at least one digit after the decimal point.".ToCharArray();
						}
					}else{
						success = false;
						errorMessages.stringx = "There must be at least one digit after the exponent symbol.".ToCharArray();
					}
				}else{
					success = false;
					errorMessages.stringx = "There must be at least one digit after the exponent symbol.".ToCharArray();
				}
			}else{
				success = false;
				errorMessages.stringx = "Expected exponent symbol.".ToCharArray();
			}
		}

		return success;
	}


	public static double GetNumberFromNumberCharacterForBase(char c, double basex){
		char [] numberTable;
		double i;
		double position;

		numberTable = GetDigitCharacterTable();
		position = 0d;

		for(i = 0d; i < basex; i = i + 1d){
			if(numberTable[(int)(i)] == c){
				position = i;
			}
		}

		return position;
	}


	public static bool CharacterIsNumberCharacterInBase(char c, double basex){
		char [] numberTable;
		double i;
		bool found;

		numberTable = GetDigitCharacterTable();
		found = false;

		for(i = 0d; i < basex; i = i + 1d){
			if(numberTable[(int)(i)] == c){
				found = true;
			}
		}

		return found;
	}


	public static double [] StringToNumberArray(char [] str){
		NumberArrayReference numberArrayReference;
		StringReference stringReference;
		double [] numbers;

		numberArrayReference = new NumberArrayReference();
		stringReference = new StringReference();

		StringToNumberArrayWithCheck(str, numberArrayReference, stringReference);

		numbers = numberArrayReference.numberArray;

		delete(numberArrayReference);
		delete(stringReference);

		return numbers;
	}


	public static bool StringToNumberArrayWithCheck(char [] str, NumberArrayReference numberArrayReference, StringReference errorMessage){
		StringReference [] numberStrings;
		double [] numbers;
		double i;
		char [] numberString, trimmedNumberString;
		bool success;
		NumberReference numberReference;

		numberStrings = SplitByString(str, ",".ToCharArray());

		numbers = new double [(int)(numberStrings.Length)];
		success = true;
		numberReference = new NumberReference();

		for(i = 0d; i < numberStrings.Length; i = i + 1d){
			numberString = numberStrings[(int)(i)].stringx;
			trimmedNumberString = Trim(numberString);
			success = CreateNumberFromDecimalStringWithCheck(trimmedNumberString, numberReference, errorMessage);
			numbers[(int)(i)] = numberReference.numberValue;

			FreeStringReference(numberStrings[(int)(i)]);
			delete(trimmedNumberString);
		}

		delete(numberStrings);
		delete(numberReference);

		numberArrayReference.numberArray = numbers;

		return success;
	}


	public static char charToLowerCase(char character){
		char toReturn;

		toReturn = character;
		if(character == 'A'){
			toReturn = 'a';
		}else if(character == 'B'){
			toReturn = 'b';
		}else if(character == 'C'){
			toReturn = 'c';
		}else if(character == 'D'){
			toReturn = 'd';
		}else if(character == 'E'){
			toReturn = 'e';
		}else if(character == 'F'){
			toReturn = 'f';
		}else if(character == 'G'){
			toReturn = 'g';
		}else if(character == 'H'){
			toReturn = 'h';
		}else if(character == 'I'){
			toReturn = 'i';
		}else if(character == 'J'){
			toReturn = 'j';
		}else if(character == 'K'){
			toReturn = 'k';
		}else if(character == 'L'){
			toReturn = 'l';
		}else if(character == 'M'){
			toReturn = 'm';
		}else if(character == 'N'){
			toReturn = 'n';
		}else if(character == 'O'){
			toReturn = 'o';
		}else if(character == 'P'){
			toReturn = 'p';
		}else if(character == 'Q'){
			toReturn = 'q';
		}else if(character == 'R'){
			toReturn = 'r';
		}else if(character == 'S'){
			toReturn = 's';
		}else if(character == 'T'){
			toReturn = 't';
		}else if(character == 'U'){
			toReturn = 'u';
		}else if(character == 'V'){
			toReturn = 'v';
		}else if(character == 'W'){
			toReturn = 'w';
		}else if(character == 'X'){
			toReturn = 'x';
		}else if(character == 'Y'){
			toReturn = 'y';
		}else if(character == 'Z'){
			toReturn = 'z';
		}

		return toReturn;
	}


	public static char charToUpperCase(char character){
		char toReturn;

		toReturn = character;
		if(character == 'a'){
			toReturn = 'A';
		}else if(character == 'b'){
			toReturn = 'B';
		}else if(character == 'c'){
			toReturn = 'C';
		}else if(character == 'd'){
			toReturn = 'D';
		}else if(character == 'e'){
			toReturn = 'E';
		}else if(character == 'f'){
			toReturn = 'F';
		}else if(character == 'g'){
			toReturn = 'G';
		}else if(character == 'h'){
			toReturn = 'H';
		}else if(character == 'i'){
			toReturn = 'I';
		}else if(character == 'j'){
			toReturn = 'J';
		}else if(character == 'k'){
			toReturn = 'K';
		}else if(character == 'l'){
			toReturn = 'L';
		}else if(character == 'm'){
			toReturn = 'M';
		}else if(character == 'n'){
			toReturn = 'N';
		}else if(character == 'o'){
			toReturn = 'O';
		}else if(character == 'p'){
			toReturn = 'P';
		}else if(character == 'q'){
			toReturn = 'Q';
		}else if(character == 'r'){
			toReturn = 'R';
		}else if(character == 's'){
			toReturn = 'S';
		}else if(character == 't'){
			toReturn = 'T';
		}else if(character == 'u'){
			toReturn = 'U';
		}else if(character == 'v'){
			toReturn = 'V';
		}else if(character == 'w'){
			toReturn = 'W';
		}else if(character == 'x'){
			toReturn = 'X';
		}else if(character == 'y'){
			toReturn = 'Y';
		}else if(character == 'z'){
			toReturn = 'Z';
		}

		return toReturn;
	}


	public static bool charIsUpperCase(char character){
		bool isUpper;

		isUpper = true;
		if(character == 'A'){
		}else if(character == 'B'){
		}else if(character == 'C'){
		}else if(character == 'D'){
		}else if(character == 'E'){
		}else if(character == 'F'){
		}else if(character == 'G'){
		}else if(character == 'H'){
		}else if(character == 'I'){
		}else if(character == 'J'){
		}else if(character == 'K'){
		}else if(character == 'L'){
		}else if(character == 'M'){
		}else if(character == 'N'){
		}else if(character == 'O'){
		}else if(character == 'P'){
		}else if(character == 'Q'){
		}else if(character == 'R'){
		}else if(character == 'S'){
		}else if(character == 'T'){
		}else if(character == 'U'){
		}else if(character == 'V'){
		}else if(character == 'W'){
		}else if(character == 'X'){
		}else if(character == 'Y'){
		}else if(character == 'Z'){
		}else{
			isUpper = false;
		}

		return isUpper;
	}


	public static bool charIsLowerCase(char character){
		bool isLower;

		isLower = true;
		if(character == 'a'){
		}else if(character == 'b'){
		}else if(character == 'c'){
		}else if(character == 'd'){
		}else if(character == 'e'){
		}else if(character == 'f'){
		}else if(character == 'g'){
		}else if(character == 'h'){
		}else if(character == 'i'){
		}else if(character == 'j'){
		}else if(character == 'k'){
		}else if(character == 'l'){
		}else if(character == 'm'){
		}else if(character == 'n'){
		}else if(character == 'o'){
		}else if(character == 'p'){
		}else if(character == 'q'){
		}else if(character == 'r'){
		}else if(character == 's'){
		}else if(character == 't'){
		}else if(character == 'u'){
		}else if(character == 'v'){
		}else if(character == 'w'){
		}else if(character == 'x'){
		}else if(character == 'y'){
		}else if(character == 'z'){
		}else{
			isLower = false;
		}

		return isLower;
	}


	public static bool charIsLetter(char character){
		return charIsUpperCase(character) || charIsLowerCase(character);
	}


	public static bool charIsNumber(char character){
		bool isNumberx;

		isNumberx = true;
		if(character == '0'){
		}else if(character == '1'){
		}else if(character == '2'){
		}else if(character == '3'){
		}else if(character == '4'){
		}else if(character == '5'){
		}else if(character == '6'){
		}else if(character == '7'){
		}else if(character == '8'){
		}else if(character == '9'){
		}else{
			isNumberx = false;
		}

		return isNumberx;
	}


	public static bool charIsWhiteSpace(char character){
		bool isWhiteSpacex;

		isWhiteSpacex = true;
		if(character == ' '){
		}else if(character == '\t'){
		}else if(character == '\n'){
		}else if(character == '\r'){
		}else{
			isWhiteSpacex = false;
		}

		return isWhiteSpacex;
	}


	public static bool charIsSymbol(char character){
		bool isSymbolx;

		isSymbolx = true;
		if(character == '!'){
		}else if(character == '\"'){
		}else if(character == '#'){
		}else if(character == '$'){
		}else if(character == '%'){
		}else if(character == '&'){
		}else if(character == '\''){
		}else if(character == '('){
		}else if(character == ')'){
		}else if(character == '*'){
		}else if(character == '+'){
		}else if(character == ','){
		}else if(character == '-'){
		}else if(character == '.'){
		}else if(character == '/'){
		}else if(character == ':'){
		}else if(character == ';'){
		}else if(character == '<'){
		}else if(character == '='){
		}else if(character == '>'){
		}else if(character == '?'){
		}else if(character == '@'){
		}else if(character == '['){
		}else if(character == '\\'){
		}else if(character == ']'){
		}else if(character == '^'){
		}else if(character == '_'){
		}else if(character == '`'){
		}else if(character == '{'){
		}else if(character == '|'){
		}else if(character == '}'){
		}else if(character == '~'){
		}else{
			isSymbolx = false;
		}

		return isSymbolx;
	}


	public static bool charCharacterIsBefore(char a, char b){
		double ad, bd;

		ad = a;
		bd = b;

		return ad < bd;
	}


	public static char charDecimalDigitToCharacter(double digit){
		char c;
		if(digit == 1d){
			c = '1';
		}else if(digit == 2d){
			c = '2';
		}else if(digit == 3d){
			c = '3';
		}else if(digit == 4d){
			c = '4';
		}else if(digit == 5d){
			c = '5';
		}else if(digit == 6d){
			c = '6';
		}else if(digit == 7d){
			c = '7';
		}else if(digit == 8d){
			c = '8';
		}else if(digit == 9d){
			c = '9';
		}else{
			c = '0';
		}
		return c;
	}


	public static double charCharacterToDecimalDigit(char c){
		double digit;

		if(c == '1'){
			digit = 1d;
		}else if(c == '2'){
			digit = 2d;
		}else if(c == '3'){
			digit = 3d;
		}else if(c == '4'){
			digit = 4d;
		}else if(c == '5'){
			digit = 5d;
		}else if(c == '6'){
			digit = 6d;
		}else if(c == '7'){
			digit = 7d;
		}else if(c == '8'){
			digit = 8d;
		}else if(c == '9'){
			digit = 9d;
		}else{
			digit = 0d;
		}

		return digit;
	}


	public static double [] arraysStringToNumberArray(char [] stringx){
		double i;
		double [] array;

		array = new double [(int)(stringx.Length)];

		for(i = 0d; i < stringx.Length; i = i + 1d){
			array[(int)(i)] = stringx[(int)(i)];
		}
		return array;
	}


	public static char [] arraysNumberArrayToString(double [] array){
		double i;
		char [] stringx;

		stringx = new char [(int)(array.Length)];

		for(i = 0d; i < array.Length; i = i + 1d){
			stringx[(int)(i)] = (char)(array[(int)(i)]);
		}
		return stringx;
	}


	public static bool arraysNumberArraysEqual(double [] a, double [] b){
		bool equal;
		double i;

		equal = true;
		if(a.Length == b.Length){
			for(i = 0d; i < a.Length && equal; i = i + 1d){
				if(a[(int)(i)] != b[(int)(i)]){
					equal = false;
				}
			}
		}else{
			equal = false;
		}

		return equal;
	}


	public static bool arraysBooleanArraysEqual(bool [] a, bool [] b){
		bool equal;
		double i;

		equal = true;
		if(a.Length == b.Length){
			for(i = 0d; i < a.Length && equal; i = i + 1d){
				if(a[(int)(i)] != b[(int)(i)]){
					equal = false;
				}
			}
		}else{
			equal = false;
		}

		return equal;
	}


	public static bool arraysStringsEqual(char [] a, char [] b){
		bool equal;
		double i;

		equal = true;
		if(a.Length == b.Length){
			for(i = 0d; i < a.Length && equal; i = i + 1d){
				if(a[(int)(i)] != b[(int)(i)]){
					equal = false;
				}
			}
		}else{
			equal = false;
		}

		return equal;
	}


	public static void arraysFillNumberArray(double [] a, double value){
		double i;

		for(i = 0d; i < a.Length; i = i + 1d){
			a[(int)(i)] = value;
		}
	}


	public static void arraysFillString(char [] a, char value){
		double i;

		for(i = 0d; i < a.Length; i = i + 1d){
			a[(int)(i)] = value;
		}
	}


	public static void arraysFillBooleanArray(bool [] a, bool value){
		double i;

		for(i = 0d; i < a.Length; i = i + 1d){
			a[(int)(i)] = value;
		}
	}


	public static bool arraysFillNumberArrayRange(double [] a, double value, double from, double to){
		double i, length;
		bool success;

		if(from >= 0d && from <= a.Length && to >= 0d && to <= a.Length && from <= to){
			length = to - from;
			for(i = 0d; i < length; i = i + 1d){
				a[(int)(from + i)] = value;
			}

			success = true;
		}else{
			success = false;
		}

		return success;
	}


	public static bool arraysFillBooleanArrayRange(bool [] a, bool value, double from, double to){
		double i, length;
		bool success;

		if(from >= 0d && from <= a.Length && to >= 0d && to <= a.Length && from <= to){
			length = to - from;
			for(i = 0d; i < length; i = i + 1d){
				a[(int)(from + i)] = value;
			}

			success = true;
		}else{
			success = false;
		}

		return success;
	}


	public static bool arraysFillStringRange(char [] a, char value, double from, double to){
		double i, length;
		bool success;

		if(from >= 0d && from <= a.Length && to >= 0d && to <= a.Length && from <= to){
			length = to - from;
			for(i = 0d; i < length; i = i + 1d){
				a[(int)(from + i)] = value;
			}

			success = true;
		}else{
			success = false;
		}

		return success;
	}


	public static double [] arraysCopyNumberArray(double [] a){
		double i;
		double [] n;

		n = new double [(int)(a.Length)];

		for(i = 0d; i < a.Length; i = i + 1d){
			n[(int)(i)] = a[(int)(i)];
		}

		return n;
	}


	public static bool [] arraysCopyBooleanArray(bool [] a){
		double i;
		bool [] n;

		n = new bool [(int)(a.Length)];

		for(i = 0d; i < a.Length; i = i + 1d){
			n[(int)(i)] = a[(int)(i)];
		}

		return n;
	}


	public static char [] arraysCopyString(char [] a){
		double i;
		char [] n;

		n = new char [(int)(a.Length)];

		for(i = 0d; i < a.Length; i = i + 1d){
			n[(int)(i)] = a[(int)(i)];
		}

		return n;
	}


	public static bool arraysCopyNumberArrayRange(double [] a, double from, double to, NumberArrayReference copyReference){
		double i, length;
		double [] n;
		bool success;

		if(from >= 0d && from <= a.Length && to >= 0d && to <= a.Length && from <= to){
			length = to - from;
			n = new double [(int)(length)];

			for(i = 0d; i < length; i = i + 1d){
				n[(int)(i)] = a[(int)(from + i)];
			}

			copyReference.numberArray = n;
			success = true;
		}else{
			success = false;
		}

		return success;
	}


	public static bool arraysCopyBooleanArrayRange(bool [] a, double from, double to, BooleanArrayReference copyReference){
		double i, length;
		bool [] n;
		bool success;

		if(from >= 0d && from <= a.Length && to >= 0d && to <= a.Length && from <= to){
			length = to - from;
			n = new bool [(int)(length)];

			for(i = 0d; i < length; i = i + 1d){
				n[(int)(i)] = a[(int)(from + i)];
			}

			copyReference.booleanArray = n;
			success = true;
		}else{
			success = false;
		}

		return success;
	}


	public static bool arraysCopyStringRange(char [] a, double from, double to, StringReference copyReference){
		double i, length;
		char [] n;
		bool success;

		if(from >= 0d && from <= a.Length && to >= 0d && to <= a.Length && from <= to){
			length = to - from;
			n = new char [(int)(length)];

			for(i = 0d; i < length; i = i + 1d){
				n[(int)(i)] = a[(int)(from + i)];
			}

			copyReference.stringx = n;
			success = true;
		}else{
			success = false;
		}

		return success;
	}


	public static bool arraysIsLastElement(double length, double index){
		return index + 1d == length;
	}


	public static double [] arraysCreateNumberArray(double length, double value){
		double [] array;

		array = new double [(int)(length)];
		arraysFillNumberArray(array, value);

		return array;
	}


	public static bool [] arraysCreateBooleanArray(double length, bool value){
		bool [] array;

		array = new bool [(int)(length)];
		arraysFillBooleanArray(array, value);

		return array;
	}


	public static char [] arraysCreateString(double length, char value){
		char [] array;

		array = new char [(int)(length)];
		arraysFillString(array, value);

		return array;
	}


	public static void arraysSwapElementsOfNumberArray(double [] A, double ai, double bi){
		double tmp;

		tmp = A[(int)(ai)];
		A[(int)(ai)] = A[(int)(bi)];
		A[(int)(bi)] = tmp;
	}


	public static void arraysSwapElementsOfStringArray(StringArrayReference A, double ai, double bi){
		StringReference tmp;

		tmp = A.stringArray[(int)(ai)];
		A.stringArray[(int)(ai)] = A.stringArray[(int)(bi)];
		A.stringArray[(int)(bi)] = tmp;
	}


	public static void arraysReverseNumberArray(double [] array){
		double i;

		for(i = 0d; i < array.Length/2d; i = i + 1d){
			arraysSwapElementsOfNumberArray(array, i, array.Length - i - 1d);
		}
	}


	public static void WriteStringToStingStream(char [] stream, NumberReference index, char [] src){
		double i;

		for(i = 0d; i < src.Length; i = i + 1d){
			stream[(int)(index.numberValue + i)] = src[(int)(i)];
		}
		index.numberValue = index.numberValue + src.Length;
	}


	public static void WriteCharacterToStingStream(char [] stream, NumberReference index, char src){
		stream[(int)(index.numberValue)] = src;
		index.numberValue = index.numberValue + 1d;
	}


	public static void WriteBooleanToStingStream(char [] stream, NumberReference index, bool src){
		if(src){
			WriteStringToStingStream(stream, index, "true".ToCharArray());
		}else{
			WriteStringToStingStream(stream, index, "false".ToCharArray());
		}
	}


	public static bool SubstringWithCheck(char [] stringx, double from, double to, StringReference stringReference){
		bool success;

		if(from >= 0d && from <= stringx.Length && to >= 0d && to <= stringx.Length && from <= to){
			stringReference.stringx = Substring(stringx, from, to);
			success = true;
		}else{
			success = false;
		}

		return success;
	}


	public static char [] Substring(char [] stringx, double from, double to){
		char [] n;
		double i, length;

		length = to - from;

		n = new char [(int)(length)];

		for(i = from; i < to; i = i + 1d){
			n[(int)(i - from)] = stringx[(int)(i)];
		}

		return n;
	}


	public static char [] AppendString(char [] s1, char [] s2){
		char [] newString;

		newString = ConcatenateString(s1, s2);

		delete(s1);

		return newString;
	}


	public static char [] ConcatenateString(char [] s1, char [] s2){
		char [] newString;
		double i;

		newString = new char [(int)(s1.Length + s2.Length)];

		for(i = 0d; i < s1.Length; i = i + 1d){
			newString[(int)(i)] = s1[(int)(i)];
		}

		for(i = 0d; i < s2.Length; i = i + 1d){
			newString[(int)(s1.Length + i)] = s2[(int)(i)];
		}

		return newString;
	}


	public static char [] AppendCharacter(char [] stringx, char c){
		char [] newString;

		newString = ConcatenateCharacter(stringx, c);

		delete(stringx);

		return newString;
	}


	public static char [] ConcatenateCharacter(char [] stringx, char c){
		char [] newString;
		double i;
		newString = new char [(int)(stringx.Length + 1d)];

		for(i = 0d; i < stringx.Length; i = i + 1d){
			newString[(int)(i)] = stringx[(int)(i)];
		}

		newString[(int)(stringx.Length)] = c;

		return newString;
	}


	public static StringReference [] SplitByCharacter(char [] toSplit, char splitBy){
		StringReference [] split;
		char [] stringToSplitBy;

		stringToSplitBy = new char [1];
		stringToSplitBy[0] = splitBy;

		split = SplitByString(toSplit, stringToSplitBy);

		delete(stringToSplitBy);

		return split;
	}


	public static bool IndexOfCharacter(char [] stringx, char character, NumberReference indexReference){
		double i;
		bool found;

		found = false;
		for(i = 0d; i < stringx.Length && !found; i = i + 1d){
			if(stringx[(int)(i)] == character){
				found = true;
				indexReference.numberValue = i;
			}
		}

		return found;
	}


	public static bool SubstringEqualsWithCheck(char [] stringx, double from, char [] substring, BooleanReference equalsReference){
		bool success;

		if(from < stringx.Length){
			success = true;
			equalsReference.booleanValue = SubstringEquals(stringx, from, substring);
		}else{
			success = false;
		}

		return success;
	}


	public static bool SubstringEquals(char [] stringx, double from, char [] substring){
		double i;
		bool equal;

		equal = true;
		if(stringx.Length - from >= substring.Length){
			for(i = 0d; i < substring.Length && equal; i = i + 1d){
				if(stringx[(int)(from + i)] != substring[(int)(i)]){
					equal = false;
				}
			}
		}else{
			equal = false;
		}

		return equal;
	}


	public static bool IndexOfString(char [] stringx, char [] substring, NumberReference indexReference){
		double i;
		bool found;

		found = false;
		for(i = 0d; i < stringx.Length - substring.Length + 1d && !found; i = i + 1d){
			if(SubstringEquals(stringx, i, substring)){
				found = true;
				indexReference.numberValue = i;
			}
		}

		return found;
	}


	public static bool ContainsCharacter(char [] stringx, char character){
		double i;
		bool found;

		found = false;
		for(i = 0d; i < stringx.Length && !found; i = i + 1d){
			if(stringx[(int)(i)] == character){
				found = true;
			}
		}

		return found;
	}


	public static bool ContainsString(char [] stringx, char [] substring){
		return IndexOfString(stringx, substring, new NumberReference());
	}


	public static void ToUpperCase(char [] stringx){
		double i;

		for(i = 0d; i < stringx.Length; i = i + 1d){
			stringx[(int)(i)] = charToUpperCase(stringx[(int)(i)]);
		}
	}


	public static void ToLowerCase(char [] stringx){
		double i;

		for(i = 0d; i < stringx.Length; i = i + 1d){
			stringx[(int)(i)] = charToLowerCase(stringx[(int)(i)]);
		}
	}


	public static bool EqualsIgnoreCase(char [] a, char [] b){
		bool equal;
		double i;

		if(a.Length == b.Length){
			equal = true;
			for(i = 0d; i < a.Length && equal; i = i + 1d){
				if(charToLowerCase(a[(int)(i)]) != charToLowerCase(b[(int)(i)])){
					equal = false;
				}
			}
		}else{
			equal = false;
		}

		return equal;
	}


	public static char [] ReplaceString(char [] stringx, char [] toReplace, char [] replaceWith){
		char [] result;
		double i, j;
		BooleanReference equalsReference;
		bool success;
		DynamicArrayCharacters da;

		da = CreateDynamicArrayCharacters();

		equalsReference = new BooleanReference();

		for(i = 0d; i < stringx.Length; ){
			success = SubstringEqualsWithCheck(stringx, i, toReplace, equalsReference);
			if(success){
				success = equalsReference.booleanValue;
			}

			if(success && toReplace.Length > 0d){
				for(j = 0d; j < replaceWith.Length; j = j + 1d){
					DynamicArrayAddCharacter(da, replaceWith[(int)(j)]);
				}
				i = i + toReplace.Length;
			}else{
				DynamicArrayAddCharacter(da, stringx[(int)(i)]);
				i = i + 1d;
			}
		}

		result = DynamicArrayCharactersToArray(da);

		FreeDynamicArrayCharacters(da);

		return result;
	}


	public static char [] ReplaceCharacterToNew(char [] stringx, char toReplace, char replaceWith){
		char [] result;
		double i;

		result = new char [(int)(stringx.Length)];

		for(i = 0d; i < stringx.Length; i = i + 1d){
			if(stringx[(int)(i)] == toReplace){
				result[(int)(i)] = replaceWith;
			}else{
				result[(int)(i)] = stringx[(int)(i)];
			}
		}

		return result;
	}


	public static void ReplaceCharacter(char [] stringx, char toReplace, char replaceWith){
		double i;

		for(i = 0d; i < stringx.Length; i = i + 1d){
			if(stringx[(int)(i)] == toReplace){
				stringx[(int)(i)] = replaceWith;
			}
		}
	}


	public static char [] Trim(char [] stringx){
		char [] result;
		double i, lastWhitespaceLocationStart, lastWhitespaceLocationEnd;
		bool firstNonWhitespaceFound;

		/* Find whitepaces at the start.*/
		lastWhitespaceLocationStart = -1d;
		firstNonWhitespaceFound = false;
		for(i = 0d; i < stringx.Length && !firstNonWhitespaceFound; i = i + 1d){
			if(charIsWhiteSpace(stringx[(int)(i)])){
				lastWhitespaceLocationStart = i;
			}else{
				firstNonWhitespaceFound = true;
			}
		}

		/* Find whitepaces at the end.*/
		lastWhitespaceLocationEnd = stringx.Length;
		firstNonWhitespaceFound = false;
		for(i = stringx.Length - 1d; i >= 0d && !firstNonWhitespaceFound; i = i - 1d){
			if(charIsWhiteSpace(stringx[(int)(i)])){
				lastWhitespaceLocationEnd = i;
			}else{
				firstNonWhitespaceFound = true;
			}
		}

		if(lastWhitespaceLocationStart < lastWhitespaceLocationEnd){
			result = Substring(stringx, lastWhitespaceLocationStart + 1d, lastWhitespaceLocationEnd);
		}else{
			result = new char [0];
		}

		return result;
	}


	public static bool StartsWith(char [] stringx, char [] start){
		bool startsWithString;

		startsWithString = false;
		if(stringx.Length >= start.Length){
			startsWithString = SubstringEquals(stringx, 0d, start);
		}

		return startsWithString;
	}


	public static bool EndsWith(char [] stringx, char [] end){
		bool endsWithString;

		endsWithString = false;
		if(stringx.Length >= end.Length){
			endsWithString = SubstringEquals(stringx, stringx.Length - end.Length, end);
		}

		return endsWithString;
	}


	public static StringReference [] SplitByString(char [] toSplit, char [] splitBy){
		StringReference [] split;
		char [] next;
		double i;
		char c;
		StringReference n;

		split = new StringReference [0];

		next = new char [0];
		for(i = 0d; i < toSplit.Length; ){
			c = toSplit[(int)(i)];

			if(SubstringEquals(toSplit, i, splitBy)){
				n = new StringReference();
				n.stringx = next;
				split = AddString(split, n);
				next = new char [0];
				i = i + splitBy.Length;
			}else{
				next = AppendCharacter(next, c);
				i = i + 1d;
			}
		}

		n = new StringReference();
		n.stringx = next;
		split = AddString(split, n);

		return split;
	}


	public static bool StringIsBefore(char [] a, char [] b){
		bool before, equal, done;
		double i;

		before = false;
		equal = true;
		done = false;

		if(a.Length == 0d && b.Length > 0d){
			before = true;
		}else{
			for(i = 0d; i < a.Length && i < b.Length && !done; i = i + 1d){
				if(a[(int)(i)] != b[(int)(i)]){
					equal = false;
				}
				if(charCharacterIsBefore(a[(int)(i)], b[(int)(i)])){
					before = true;
				}
				if(charCharacterIsBefore(b[(int)(i)], a[(int)(i)])){
					done = true;
				}
			}

			if(equal){
				if(a.Length < b.Length){
					before = true;
				}
			}
		}

		return before;
	}


	public static char [] JoinStringsWithSeparator(StringReference [] strings, char [] separator){
		char [] result, stringx;
		double length, i;
		NumberReference index;

		index = CreateNumberReference(0d);

		length = 0d;
		for(i = 0d; i < strings.Length; i = i + 1d){
			length = length + strings[(int)(i)].stringx.Length;
		}
		length = length + (strings.Length - 1d)*separator.Length;

		result = new char [(int)(length)];

		for(i = 0d; i < strings.Length; i = i + 1d){
			stringx = strings[(int)(i)].stringx;
			WriteStringToStingStream(result, index, stringx);
			if(i + 1d < strings.Length){
				WriteStringToStingStream(result, index, separator);
			}
		}

		delete(index);

		return result;
	}


	public static char [] JoinStrings(StringReference [] strings){
		char [] result, stringx;
		double length, i;
		NumberReference index;

		index = CreateNumberReference(0d);

		length = 0d;
		for(i = 0d; i < strings.Length; i = i + 1d){
			length = length + strings[(int)(i)].stringx.Length;
		}

		result = new char [(int)(length)];

		for(i = 0d; i < strings.Length; i = i + 1d){
			stringx = strings[(int)(i)].stringx;
			WriteStringToStingStream(result, index, stringx);
		}

		delete(index);

		return result;
	}


	public static double [] AddNumber(double [] list, double a){
		double [] newlist;
		double i;

		newlist = new double [(int)(list.Length + 1d)];
		for(i = 0d; i < list.Length; i = i + 1d){
			newlist[(int)(i)] = list[(int)(i)];
		}
		newlist[(int)(list.Length)] = a;
		
		delete(list);
		
		return newlist;
	}


	public static void AddNumberRef(NumberArrayReference list, double i){
		list.numberArray = AddNumber(list.numberArray, i);
	}


	public static double [] RemoveNumber(double [] list, double n){
		double [] newlist;
		double i;

		newlist = new double [(int)(list.Length - 1d)];

		if(n >= 0d && n < list.Length){
			for(i = 0d; i < list.Length; i = i + 1d){
				if(i < n){
					newlist[(int)(i)] = list[(int)(i)];
				}
				if(i > n){
					newlist[(int)(i - 1d)] = list[(int)(i)];
				}
			}

			delete(list);
		}else{
			delete(newlist);
		}
		
		return newlist;
	}


	public static double GetNumberRef(NumberArrayReference list, double i){
		return list.numberArray[(int)(i)];
	}


	public static void RemoveNumberRef(NumberArrayReference list, double i){
		list.numberArray = RemoveNumber(list.numberArray, i);
	}


	public static StringReference [] AddString(StringReference [] list, StringReference a){
		StringReference [] newlist;
		double i;

		newlist = new StringReference [(int)(list.Length + 1d)];

		for(i = 0d; i < list.Length; i = i + 1d){
			newlist[(int)(i)] = list[(int)(i)];
		}
		newlist[(int)(list.Length)] = a;
		
		delete(list);
		
		return newlist;
	}


	public static void AddStringRef(StringArrayReference list, StringReference i){
		list.stringArray = AddString(list.stringArray, i);
	}


	public static StringReference [] RemoveString(StringReference [] list, double n){
		StringReference [] newlist;
		double i;

		newlist = new StringReference [(int)(list.Length - 1d)];

		if(n >= 0d && n < list.Length){
			for(i = 0d; i < list.Length; i = i + 1d){
				if(i < n){
					newlist[(int)(i)] = list[(int)(i)];
				}
				if(i > n){
					newlist[(int)(i - 1d)] = list[(int)(i)];
				}
			}

			delete(list);
		}else{
			delete(newlist);
		}
		
		return newlist;
	}


	public static StringReference GetStringRef(StringArrayReference list, double i){
		return list.stringArray[(int)(i)];
	}


	public static void RemoveStringRef(StringArrayReference list, double i){
		list.stringArray = RemoveString(list.stringArray, i);
	}


	public static DynamicArrayCharacters CreateDynamicArrayCharacters(){
		DynamicArrayCharacters da;

		da = new DynamicArrayCharacters();
		da.array = new char [10];
		da.length = 0d;

		return da;
	}


	public static DynamicArrayCharacters CreateDynamicArrayCharactersWithInitialCapacity(double capacity){
		DynamicArrayCharacters da;

		da = new DynamicArrayCharacters();
		da.array = new char [(int)(capacity)];
		da.length = 0d;

		return da;
	}


	public static void DynamicArrayAddCharacter(DynamicArrayCharacters da, char value){
		if(da.length == da.array.Length){
			DynamicArrayCharactersIncreaseSize(da);
		}

		da.array[(int)(da.length)] = value;
		da.length = da.length + 1d;
	}


	public static void DynamicArrayCharactersIncreaseSize(DynamicArrayCharacters da){
		double newLength, i;
		char [] newArray;

		newLength = (double)Round(da.array.Length*3d/2d);
		newArray = new char [(int)(newLength)];

		for(i = 0d; i < da.array.Length; i = i + 1d){
			newArray[(int)(i)] = da.array[(int)(i)];
		}

		delete(da.array);

		da.array = newArray;
	}


	public static bool DynamicArrayCharactersDecreaseSizeNecessary(DynamicArrayCharacters da){
		bool needsDecrease;

		needsDecrease = false;

		if(da.length > 10d){
			needsDecrease = da.length <= (double)Round(da.array.Length*2d/3d);
		}

		return needsDecrease;
	}


	public static void DynamicArrayCharactersDecreaseSize(DynamicArrayCharacters da){
		double newLength, i;
		char [] newArray;

		newLength = (double)Round(da.array.Length*2d/3d);
		newArray = new char [(int)(newLength)];

		for(i = 0d; i < newLength; i = i + 1d){
			newArray[(int)(i)] = da.array[(int)(i)];
		}

		delete(da.array);

		da.array = newArray;
	}


	public static double DynamicArrayCharactersIndex(DynamicArrayCharacters da, double index){
		return da.array[(int)(index)];
	}


	public static double DynamicArrayCharactersLength(DynamicArrayCharacters da){
		return da.length;
	}


	public static void DynamicArrayInsertCharacter(DynamicArrayCharacters da, double index, char value){
		double i;

		if(da.length == da.array.Length){
			DynamicArrayCharactersIncreaseSize(da);
		}

		for(i = da.length; i > index; i = i - 1d){
			da.array[(int)(i)] = da.array[(int)(i - 1d)];
		}

		da.array[(int)(index)] = value;

		da.length = da.length + 1d;
	}


	public static bool DynamicArrayCharacterSet(DynamicArrayCharacters da, double index, char value){
		bool success;

		if(index < da.length){
			da.array[(int)(index)] = value;
			success = true;
		}else{
			success = false;
		}

		return success;
	}


	public static void DynamicArrayRemoveCharacter(DynamicArrayCharacters da, double index){
		double i;

		for(i = index; i < da.length - 1d; i = i + 1d){
			da.array[(int)(i)] = da.array[(int)(i + 1d)];
		}

		da.length = da.length - 1d;

		if(DynamicArrayCharactersDecreaseSizeNecessary(da)){
			DynamicArrayCharactersDecreaseSize(da);
		}
	}


	public static void FreeDynamicArrayCharacters(DynamicArrayCharacters da){
		delete(da.array);
		delete(da);
	}


	public static char [] DynamicArrayCharactersToArray(DynamicArrayCharacters da){
		char [] array;
		double i;

		array = new char [(int)(da.length)];

		for(i = 0d; i < da.length; i = i + 1d){
			array[(int)(i)] = da.array[(int)(i)];
		}

		return array;
	}


	public static DynamicArrayCharacters ArrayToDynamicArrayCharactersWithOptimalSize(char [] array){
		DynamicArrayCharacters da;
		double i;
		double c, n, newCapacity;

		c = array.Length;
		n = (Log(c) - 1d)/Log(3d/2d);
		newCapacity = Floor(n) + 1d;

		da = CreateDynamicArrayCharactersWithInitialCapacity(newCapacity);

		for(i = 0d; i < array.Length; i = i + 1d){
			da.array[(int)(i)] = array[(int)(i)];
		}

		return da;
	}


	public static DynamicArrayCharacters ArrayToDynamicArrayCharacters(char [] array){
		DynamicArrayCharacters da;

		da = new DynamicArrayCharacters();
		da.array = arraysCopyString(array);
		da.length = array.Length;

		return da;
	}


	public static bool DynamicArrayCharactersEqual(DynamicArrayCharacters a, DynamicArrayCharacters b){
		bool equal;
		double i;

		equal = true;
		if(a.length == b.length){
			for(i = 0d; i < a.length && equal; i = i + 1d){
				if(a.array[(int)(i)] != b.array[(int)(i)]){
					equal = false;
				}
			}
		}else{
			equal = false;
		}

		return equal;
	}


	public static LinkedListCharacters DynamicArrayCharactersToLinkedList(DynamicArrayCharacters da){
		LinkedListCharacters ll;
		double i;

		ll = CreateLinkedListCharacter();

		for(i = 0d; i < da.length; i = i + 1d){
			LinkedListAddCharacter(ll, da.array[(int)(i)]);
		}

		return ll;
	}


	public static DynamicArrayCharacters LinkedListToDynamicArrayCharacters(LinkedListCharacters ll){
		DynamicArrayCharacters da;
		double i;
		LinkedListNodeCharacters node;

		node = ll.first;

		da = new DynamicArrayCharacters();
		da.length = LinkedListCharactersLength(ll);

		da.array = new char [(int)(da.length)];

		for(i = 0d; i < da.length; i = i + 1d){
			da.array[(int)(i)] = node.value;
			node = node.next;
		}

		return da;
	}


	public static bool [] AddBoolean(bool [] list, bool a){
		bool [] newlist;
		double i;

		newlist = new bool [(int)(list.Length + 1d)];
		for(i = 0d; i < list.Length; i = i + 1d){
			newlist[(int)(i)] = list[(int)(i)];
		}
		newlist[(int)(list.Length)] = a;
		
		delete(list);
		
		return newlist;
	}


	public static void AddBooleanRef(BooleanArrayReference list, bool i){
		list.booleanArray = AddBoolean(list.booleanArray, i);
	}


	public static bool [] RemoveBoolean(bool [] list, double n){
		bool [] newlist;
		double i;

		newlist = new bool [(int)(list.Length - 1d)];

		if(n >= 0d && n < list.Length){
			for(i = 0d; i < list.Length; i = i + 1d){
				if(i < n){
					newlist[(int)(i)] = list[(int)(i)];
				}
				if(i > n){
					newlist[(int)(i - 1d)] = list[(int)(i)];
				}
			}

			delete(list);
		}else{
			delete(newlist);
		}
		
		return newlist;
	}


	public static bool GetBooleanRef(BooleanArrayReference list, double i){
		return list.booleanArray[(int)(i)];
	}


	public static void RemoveDecimalRef(BooleanArrayReference list, double i){
		list.booleanArray = RemoveBoolean(list.booleanArray, i);
	}


	public static LinkedListStrings CreateLinkedListString(){
		LinkedListStrings ll;

		ll = new LinkedListStrings();
		ll.first = new LinkedListNodeStrings();
		ll.last = ll.first;
		ll.last.end = true;

		return ll;
	}


	public static void LinkedListAddString(LinkedListStrings ll, char [] value){
		ll.last.end = false;
		ll.last.value = value;
		ll.last.next = new LinkedListNodeStrings();
		ll.last.next.end = true;
		ll.last = ll.last.next;
	}


	public static StringReference [] LinkedListStringsToArray(LinkedListStrings ll){
		StringReference [] array;
		double length, i;
		LinkedListNodeStrings node;

		node = ll.first;

		length = LinkedListStringsLength(ll);

		array = new StringReference [(int)(length)];

		for(i = 0d; i < length; i = i + 1d){
			array[(int)(i)] = new StringReference();
			array[(int)(i)].stringx = node.value;
			node = node.next;
		}

		return array;
	}


	public static double LinkedListStringsLength(LinkedListStrings ll){
		double l;
		LinkedListNodeStrings node;

		l = 0d;
		node = ll.first;
		for(; !node.end; ){
			node = node.next;
			l = l + 1d;
		}

		return l;
	}


	public static void FreeLinkedListString(LinkedListStrings ll){
		LinkedListNodeStrings node, prev;

		node = ll.first;

		for(; !node.end; ){
			prev = node;
			node = node.next;
			delete(prev);
		}

		delete(node);
	}


	public static LinkedListNumbers CreateLinkedListNumbers(){
		LinkedListNumbers ll;

		ll = new LinkedListNumbers();
		ll.first = new LinkedListNodeNumbers();
		ll.last = ll.first;
		ll.last.end = true;

		return ll;
	}


	public static LinkedListNumbers [] CreateLinkedListNumbersArray(double length){
		LinkedListNumbers [] lls;
		double i;

		lls = new LinkedListNumbers [(int)(length)];
		for(i = 0d; i < lls.Length; i = i + 1d){
			lls[(int)(i)] = CreateLinkedListNumbers();
		}

		return lls;
	}


	public static void LinkedListAddNumber(LinkedListNumbers ll, double value){
		ll.last.end = false;
		ll.last.value = value;
		ll.last.next = new LinkedListNodeNumbers();
		ll.last.next.end = true;
		ll.last = ll.last.next;
	}


	public static double LinkedListNumbersLength(LinkedListNumbers ll){
		double l;
		LinkedListNodeNumbers node;

		l = 0d;
		node = ll.first;
		for(; !node.end; ){
			node = node.next;
			l = l + 1d;
		}

		return l;
	}


	public static double LinkedListNumbersIndex(LinkedListNumbers ll, double index){
		double i;
		LinkedListNodeNumbers node;

		node = ll.first;
		for(i = 0d; i < index; i = i + 1d){
			node = node.next;
		}

		return node.value;
	}


	public static void LinkedListInsertNumber(LinkedListNumbers ll, double index, double value){
		double i;
		LinkedListNodeNumbers node, tmp;

		if(index == 0d){
			tmp = ll.first;
			ll.first = new LinkedListNodeNumbers();
			ll.first.next = tmp;
			ll.first.value = value;
			ll.first.end = false;
		}else{
			node = ll.first;
			for(i = 0d; i < index - 1d; i = i + 1d){
				node = node.next;
			}

			tmp = node.next;
			node.next = new LinkedListNodeNumbers();
			node.next.next = tmp;
			node.next.value = value;
			node.next.end = false;
		}
	}


	public static void LinkedListSet(LinkedListNumbers ll, double index, double value){
		double i;
		LinkedListNodeNumbers node;

		node = ll.first;
		for(i = 0d; i < index; i = i + 1d){
			node = node.next;
		}

		node.next.value = value;
	}


	public static void LinkedListRemoveNumber(LinkedListNumbers ll, double index){
		double i;
		LinkedListNodeNumbers node, prev;

		node = ll.first;
		prev = ll.first;

		for(i = 0d; i < index; i = i + 1d){
			prev = node;
			node = node.next;
		}

		if(index == 0d){
			ll.first = prev.next;
		}
		if(!prev.next.end){
			prev.next = prev.next.next;
		}
	}


	public static void FreeLinkedListNumbers(LinkedListNumbers ll){
		LinkedListNodeNumbers node, prev;

		node = ll.first;

		for(; !node.end; ){
			prev = node;
			node = node.next;
			delete(prev);
		}

		delete(node);
	}


	public static void FreeLinkedListNumbersArray(LinkedListNumbers [] lls){
		double i;

		for(i = 0d; i < lls.Length; i = i + 1d){
			FreeLinkedListNumbers(lls[(int)(i)]);
		}
		delete(lls);
	}


	public static double [] LinkedListNumbersToArray(LinkedListNumbers ll){
		double [] array;
		double length, i;
		LinkedListNodeNumbers node;

		node = ll.first;

		length = LinkedListNumbersLength(ll);

		array = new double [(int)(length)];

		for(i = 0d; i < length; i = i + 1d){
			array[(int)(i)] = node.value;
			node = node.next;
		}

		return array;
	}


	public static LinkedListNumbers ArrayToLinkedListNumbers(double [] array){
		LinkedListNumbers ll;
		double i;

		ll = CreateLinkedListNumbers();

		for(i = 0d; i < array.Length; i = i + 1d){
			LinkedListAddNumber(ll, array[(int)(i)]);
		}

		return ll;
	}


	public static bool LinkedListNumbersEqual(LinkedListNumbers a, LinkedListNumbers b){
		bool equal, done;
		LinkedListNodeNumbers an, bn;

		an = a.first;
		bn = b.first;

		equal = true;
		done = false;
		for(; equal && !done; ){
			if(an.end == bn.end){
				if(an.end){
					done = true;
				}else if(an.value == bn.value){
					an = an.next;
					bn = bn.next;
				}else{
					equal = false;
				}
			}else{
				equal = false;
			}
		}

		return equal;
	}


	public static LinkedListCharacters CreateLinkedListCharacter(){
		LinkedListCharacters ll;

		ll = new LinkedListCharacters();
		ll.first = new LinkedListNodeCharacters();
		ll.last = ll.first;
		ll.last.end = true;

		return ll;
	}


	public static void LinkedListAddCharacter(LinkedListCharacters ll, char value){
		ll.last.end = false;
		ll.last.value = value;
		ll.last.next = new LinkedListNodeCharacters();
		ll.last.next.end = true;
		ll.last = ll.last.next;
	}


	public static char [] LinkedListCharactersToArray(LinkedListCharacters ll){
		char [] array;
		double length, i;
		LinkedListNodeCharacters node;

		node = ll.first;

		length = LinkedListCharactersLength(ll);

		array = new char [(int)(length)];

		for(i = 0d; i < length; i = i + 1d){
			array[(int)(i)] = node.value;
			node = node.next;
		}

		return array;
	}


	public static double LinkedListCharactersLength(LinkedListCharacters ll){
		double l;
		LinkedListNodeCharacters node;

		l = 0d;
		node = ll.first;
		for(; !node.end; ){
			node = node.next;
			l = l + 1d;
		}

		return l;
	}


	public static void FreeLinkedListCharacter(LinkedListCharacters ll){
		LinkedListNodeCharacters node, prev;

		node = ll.first;

		for(; !node.end; ){
			prev = node;
			node = node.next;
			delete(prev);
		}

		delete(node);
	}


	public static void LinkedListCharactersAddString(LinkedListCharacters ll, char [] str){
		double i;

		for(i = 0d; i < str.Length; i = i + 1d){
			LinkedListAddCharacter(ll, str[(int)(i)]);
		}
	}


	public static DynamicArrayNumbers CreateDynamicArrayNumbers(){
		DynamicArrayNumbers da;

		da = new DynamicArrayNumbers();
		da.array = new double [10];
		da.length = 0d;

		return da;
	}


	public static DynamicArrayNumbers CreateDynamicArrayNumbersWithInitialCapacity(double capacity){
		DynamicArrayNumbers da;

		da = new DynamicArrayNumbers();
		da.array = new double [(int)(capacity)];
		da.length = 0d;

		return da;
	}


	public static void DynamicArrayAddNumber(DynamicArrayNumbers da, double value){
		if(da.length == da.array.Length){
			DynamicArrayNumbersIncreaseSize(da);
		}

		da.array[(int)(da.length)] = value;
		da.length = da.length + 1d;
	}


	public static void DynamicArrayNumbersIncreaseSize(DynamicArrayNumbers da){
		double newLength, i;
		double [] newArray;

		newLength = (double)Round(da.array.Length*3d/2d);
		newArray = new double [(int)(newLength)];

		for(i = 0d; i < da.array.Length; i = i + 1d){
			newArray[(int)(i)] = da.array[(int)(i)];
		}

		delete(da.array);

		da.array = newArray;
	}


	public static bool DynamicArrayNumbersDecreaseSizeNecessary(DynamicArrayNumbers da){
		bool needsDecrease;

		needsDecrease = false;

		if(da.length > 10d){
			needsDecrease = da.length <= (double)Round(da.array.Length*2d/3d);
		}

		return needsDecrease;
	}


	public static void DynamicArrayNumbersDecreaseSize(DynamicArrayNumbers da){
		double newLength, i;
		double [] newArray;

		newLength = (double)Round(da.array.Length*2d/3d);
		newArray = new double [(int)(newLength)];

		for(i = 0d; i < newLength; i = i + 1d){
			newArray[(int)(i)] = da.array[(int)(i)];
		}

		delete(da.array);

		da.array = newArray;
	}


	public static double DynamicArrayNumbersIndex(DynamicArrayNumbers da, double index){
		return da.array[(int)(index)];
	}


	public static double DynamicArrayNumbersLength(DynamicArrayNumbers da){
		return da.length;
	}


	public static void DynamicArrayInsertNumber(DynamicArrayNumbers da, double index, double value){
		double i;

		if(da.length == da.array.Length){
			DynamicArrayNumbersIncreaseSize(da);
		}

		for(i = da.length; i > index; i = i - 1d){
			da.array[(int)(i)] = da.array[(int)(i - 1d)];
		}

		da.array[(int)(index)] = value;

		da.length = da.length + 1d;
	}


	public static bool DynamicArrayNumberSet(DynamicArrayNumbers da, double index, double value){
		bool success;

		if(index < da.length){
			da.array[(int)(index)] = value;
			success = true;
		}else{
			success = false;
		}

		return success;
	}


	public static void DynamicArrayRemoveNumber(DynamicArrayNumbers da, double index){
		double i;

		for(i = index; i < da.length - 1d; i = i + 1d){
			da.array[(int)(i)] = da.array[(int)(i + 1d)];
		}

		da.length = da.length - 1d;

		if(DynamicArrayNumbersDecreaseSizeNecessary(da)){
			DynamicArrayNumbersDecreaseSize(da);
		}
	}


	public static void FreeDynamicArrayNumbers(DynamicArrayNumbers da){
		delete(da.array);
		delete(da);
	}


	public static double [] DynamicArrayNumbersToArray(DynamicArrayNumbers da){
		double [] array;
		double i;

		array = new double [(int)(da.length)];

		for(i = 0d; i < da.length; i = i + 1d){
			array[(int)(i)] = da.array[(int)(i)];
		}

		return array;
	}


	public static DynamicArrayNumbers ArrayToDynamicArrayNumbersWithOptimalSize(double [] array){
		DynamicArrayNumbers da;
		double i;
		double c, n, newCapacity;

		/*
         c = 10*(3/2)^n
         log(c) = log(10*(3/2)^n)
         log(c) = log(10) + log((3/2)^n)
         log(c) = 1 + log((3/2)^n)
         log(c) - 1 = log((3/2)^n)
         log(c) - 1 = n*log(3/2)
         n = (log(c) - 1)/log(3/2)
        */
		c = array.Length;
		n = (Log(c) - 1d)/Log(3d/2d);
		newCapacity = Floor(n) + 1d;

		da = CreateDynamicArrayNumbersWithInitialCapacity(newCapacity);

		for(i = 0d; i < array.Length; i = i + 1d){
			da.array[(int)(i)] = array[(int)(i)];
		}

		return da;
	}


	public static DynamicArrayNumbers ArrayToDynamicArrayNumbers(double [] array){
		DynamicArrayNumbers da;

		da = new DynamicArrayNumbers();
		da.array = arraysCopyNumberArray(array);
		da.length = array.Length;

		return da;
	}


	public static bool DynamicArrayNumbersEqual(DynamicArrayNumbers a, DynamicArrayNumbers b){
		bool equal;
		double i;

		equal = true;
		if(a.length == b.length){
			for(i = 0d; i < a.length && equal; i = i + 1d){
				if(a.array[(int)(i)] != b.array[(int)(i)]){
					equal = false;
				}
			}
		}else{
			equal = false;
		}

		return equal;
	}


	public static LinkedListNumbers DynamicArrayNumbersToLinkedList(DynamicArrayNumbers da){
		LinkedListNumbers ll;
		double i;

		ll = CreateLinkedListNumbers();

		for(i = 0d; i < da.length; i = i + 1d){
			LinkedListAddNumber(ll, da.array[(int)(i)]);
		}

		return ll;
	}


	public static DynamicArrayNumbers LinkedListToDynamicArrayNumbers(LinkedListNumbers ll){
		DynamicArrayNumbers da;
		double i;
		LinkedListNodeNumbers node;

		node = ll.first;

		da = new DynamicArrayNumbers();
		da.length = LinkedListNumbersLength(ll);

		da.array = new double [(int)(da.length)];

		for(i = 0d; i < da.length; i = i + 1d){
			da.array[(int)(i)] = node.value;
			node = node.next;
		}

		return da;
	}


	public static double DynamicArrayNumbersIndexOf(DynamicArrayNumbers arr, double n, BooleanReference foundReference){
		bool found;
		double i;

		found = false;
		for(i = 0d; i < arr.length && !found; i = i + 1d){
			if(arr.array[(int)(i)] == n){
				found = true;
			}
		}
		if(!found){
			i = -1d;
		}else{
			i = i - 1d;
		}

		foundReference.booleanValue = found;

		return i;
	}


	public static bool DynamicArrayNumbersIsInArray(DynamicArrayNumbers arr, double n){
		bool found;
		double i;

		found = false;
		for(i = 0d; i < arr.length && !found; i = i + 1d){
			if(arr.array[(int)(i)] == n){
				found = true;
			}
		}

		return found;
	}


	public static char [] AddCharacter(char [] list, char a){
		char [] newlist;
		double i;

		newlist = new char [(int)(list.Length + 1d)];
		for(i = 0d; i < list.Length; i = i + 1d){
			newlist[(int)(i)] = list[(int)(i)];
		}
		newlist[(int)(list.Length)] = a;
		
		delete(list);
		
		return newlist;
	}


	public static void AddCharacterRef(StringReference list, char i){
		list.stringx = AddCharacter(list.stringx, i);
	}


	public static char [] RemoveCharacter(char [] list, double n){
		char [] newlist;
		double i;

		newlist = new char [(int)(list.Length - 1d)];

		if(n >= 0d && n < list.Length){
			for(i = 0d; i < list.Length; i = i + 1d){
				if(i < n){
					newlist[(int)(i)] = list[(int)(i)];
				}
				if(i > n){
					newlist[(int)(i - 1d)] = list[(int)(i)];
				}
			}

			delete(list);
		}else{
			delete(newlist);
		}

		return newlist;
	}


	public static char GetCharacterRef(StringReference list, double i){
		return list.stringx[(int)(i)];
	}


	public static void RemoveCharacterRef(StringReference list, double i){
		list.stringx = RemoveCharacter(list.stringx, i);
	}


	public static void AssertFalse(bool b, NumberReference failures){
		if(b){
			failures.numberValue = failures.numberValue + 1d;
		}
	}


	public static void AssertTrue(bool b, NumberReference failures){
		if(!b){
			failures.numberValue = failures.numberValue + 1d;
		}
	}


	public static void AssertEquals(double a, double b, NumberReference failures){
		if(a != b){
			failures.numberValue = failures.numberValue + 1d;
		}
	}


	public static void AssertBooleansEqual(bool a, bool b, NumberReference failures){
		if(a != b){
			failures.numberValue = failures.numberValue + 1d;
		}
	}


	public static void AssertCharactersEqual(char a, char b, NumberReference failures){
		if(a != b){
			failures.numberValue = failures.numberValue + 1d;
		}
	}


	public static void AssertStringEquals(char [] a, char [] b, NumberReference failures){
		if(!arraysStringsEqual(a, b)){
			failures.numberValue = failures.numberValue + 1d;
		}
	}


	public static void AssertNumberArraysEqual(double [] a, double [] b, NumberReference failures){
		double i;

		if(a.Length == b.Length){
			for(i = 0d; i < a.Length; i = i + 1d){
				AssertEquals(a[(int)(i)], b[(int)(i)], failures);
			}
		}else{
			failures.numberValue = failures.numberValue + 1d;
		}
	}


	public static void AssertBooleanArraysEqual(bool [] a, bool [] b, NumberReference failures){
		double i;

		if(a.Length == b.Length){
			for(i = 0d; i < a.Length; i = i + 1d){
				AssertBooleansEqual(a[(int)(i)], b[(int)(i)], failures);
			}
		}else{
			failures.numberValue = failures.numberValue + 1d;
		}
	}


	public static void AssertStringArraysEqual(StringReference [] a, StringReference [] b, NumberReference failures){
		double i;

		if(a.Length == b.Length){
			for(i = 0d; i < a.Length; i = i + 1d){
				AssertStringEquals(a[(int)(i)].stringx, b[(int)(i)].stringx, failures);
			}
		}else{
			failures.numberValue = failures.numberValue + 1d;
		}
	}


	public static double m18Negate(double x){
		return -x;
	}


	public static double m18Positive(double x){
		return +x;
	}


	public static double m18Factorial(double x){
		double i, f;

		f = 1d;

		for(i = 2d; i <= x; i = i + 1d){
			f = f*i;
		}

		return f;
	}


	public static double m18Round(double x){
		return Floor(x + 0.5);
	}


	public static double m18BankersRound(double x){
		double r;

		if(m18Absolute(x - m18Truncate(x)) == 0.5){
			if(!m18DivisibleBy(m18Round(x), 2d)){
				r = m18Round(x) - 1d;
			}else{
				r = m18Round(x);
			}
		}else{
			r = m18Round(x);
		}

		return r;
	}


	public static double m18Ceil(double x){
		return Ceiling(x);
	}


	public static double m18Floor(double x){
		return Floor(x);
	}


	public static double m18Truncate(double x){
		double t;

		if(x >= 0d){
			t = Floor(x);
		}else{
			t = Ceiling(x);
		}

		return t;
	}


	public static double m18Absolute(double x){
		return Abs(x);
	}


	public static double m18Logarithm(double x){
		return Log10(x);
	}


	public static double m18NaturalLogarithm(double x){
		return Log(x);
	}


	public static double m18Sin(double x){
		return Sin(x);
	}


	public static double m18Cos(double x){
		return Cos(x);
	}


	public static double m18Tan(double x){
		return Tan(x);
	}


	public static double m18Asin(double x){
		return Asin(x);
	}


	public static double m18Acos(double x){
		return Acos(x);
	}


	public static double m18Atan(double x){
		return Atan(x);
	}


	public static double m18Atan2(double y, double x){
		double a;

		/* Atan2 is an invalid operation when x = 0 and y = 0, but this method does not return errors.*/
		a = 0d;

		if(x > 0d){
			a = m18Atan(y/x);
		}else if(x < 0d && y >= 0d){
			a = m18Atan(y/x) + PI;
		}else if(x < 0d && y < 0d){
			a = m18Atan(y/x) - PI;
		}else if(x == 0d && y > 0d){
			a = PI/2d;
		}else if(x == 0d && y < 0d){
			a = -PI/2d;
		}

		return a;
	}


	public static double m18Squareroot(double x){
		return Sqrt(x);
	}


	public static double m18Exp(double x){
		return Exp(x);
	}


	public static bool m18DivisibleBy(double a, double b){
		return ((a%b) == 0d);
	}


	public static double m18Combinations(double n, double k){
		double i, j, c;

		c = 1d;
		j = 1d;
		i = n - k + 1d;

		for(; i <= n; ){
			c = c*i;
			c = c/j;

			i = i + 1d;
			j = j + 1d;
		}

		return c;
	}


	public static double m18Permutations(double n, double k){
		double i, c;

		c = 1d;

		for(i = n - k + 1d; i <= n; i = i + 1d){
			c = c*i;
		}

		return c;
	}


	public static bool m18EpsilonCompare(double a, double b, double epsilon){
		return Abs(a - b) < epsilon;
	}


	public static double m18GreatestCommonDivisor(double a, double b){
		double t;

		for(; b != 0d; ){
			t = b;
			b = a%b;
			a = t;
		}

		return a;
	}


	public static double m18GCDWithSubtraction(double a, double b){
		double g;

		if(a == 0d){
			g = b;
		}else{
			for(; b != 0d; ){
				if(a > b){
					a = a - b;
				}else{
					b = b - a;
				}
			}

			g = a;
		}

		return g;
	}


	public static bool m18IsInteger(double a){
		return (a - Floor(a)) == 0d;
	}


	public static bool m18GreatestCommonDivisorWithCheck(double a, double b, NumberReference gcdReference){
		bool success;
		double gcd;

		if(m18IsInteger(a) && m18IsInteger(b)){
			gcd = m18GreatestCommonDivisor(a, b);
			gcdReference.numberValue = gcd;
			success = true;
		}else{
			success = false;
		}

		return success;
	}


	public static double m18LeastCommonMultiple(double a, double b){
		double lcm;

		if(a > 0d && b > 0d){
			lcm = Abs(a*b)/m18GreatestCommonDivisor(a, b);
		}else{
			lcm = 0d;
		}

		return lcm;
	}


	public static double m18Sign(double a){
		double s;

		if(a > 0d){
			s = 1d;
		}else if(a < 0d){
			s = -1d;
		}else{
			s = 0d;
		}

		return s;
	}


	public static double m18Max(double a, double b){
		return Max(a, b);
	}


	public static double m18Min(double a, double b){
		return Min(a, b);
	}


	public static double m18Power(double a, double b){
		return Pow(a, b);
	}


	public static double m18Gamma(double x){
		return m18LanczosApproximation(x);
	}


	public static double m18LogGamma(double x){
		return Log(m18Gamma(x));
	}


	public static double m18LanczosApproximation(double z){
		double [] p;
		double i, y, t, x;

		p = new double [8];
		p[0] = 676.5203681218851;
		p[1] = -1259.1392167224028;
		p[2] = 771.32342877765313;
		p[3] = -176.61502916214059;
		p[4] = 12.507343278686905;
		p[5] = -0.13857109526572012;
		p[6] = 9.9843695780195716e-6;
		p[7] = 1.5056327351493116e-7;

		if(z < 0.5){
			y = PI/(Sin(PI*z)*m18LanczosApproximation(1d - z));
		}else{
			z = z - 1d;
			x = 0.99999999999980993;
			for(i = 0d; i < p.Length; i = i + 1d){
				x = x + p[(int)(i)]/(z + i + 1d);
			}
			t = z + p.Length - 0.5;
			y = Sqrt(2d*PI)*Pow(t, z + 0.5)*Exp(-t)*x;
		}

		return y;
	}


	public static double m18Beta(double x, double y){
		return m18Gamma(x)*m18Gamma(y)/m18Gamma(x + y);
	}


	public static double m18Sinh(double x){
		return (Exp(x) - Exp(-x))/2d;
	}


	public static double m18Cosh(double x){
		return (Exp(x) + Exp(-x))/2d;
	}


	public static double m18Tanh(double x){
		return m18Sinh(x)/m18Cosh(x);
	}


	public static double m18Cot(double x){
		return 1d/Tan(x);
	}


	public static double m18Sec(double x){
		return 1d/Cos(x);
	}


	public static double m18Csc(double x){
		return 1d/Sin(x);
	}


	public static double m18Coth(double x){
		return m18Cosh(x)/m18Sinh(x);
	}


	public static double m18Sech(double x){
		return 1d/m18Cosh(x);
	}


	public static double m18Csch(double x){
		return 1d/m18Sinh(x);
	}


	public static double m18Error(double x){
		double y, t, tau, c1, c2, c3, c4, c5, c6, c7, c8, c9, c10;

		if(x == 0d){
			y = 0d;
		}else if(x < 0d){
			y = -m18Error(-x);
		}else{
			c1 = -1.26551223;
			c2 = +1.00002368;
			c3 = +0.37409196;
			c4 = +0.09678418;
			c5 = -0.18628806;
			c6 = +0.27886807;
			c7 = -1.13520398;
			c8 = +1.48851587;
			c9 = -0.82215223;
			c10 = +0.17087277;

			t = 1d/(1d + 0.5*Abs(x));

			tau = t*Exp(-Pow(x, 2d) + c1 + t*(c2 + t*(c3 + t*(c4 + t*(c5 + t*(c6 + t*(c7 + t*(c8 + t*(c9 + t*c10)))))))));

			y = 1d - tau;
		}

		return y;
	}


	public static double m18ErrorInverse(double x){
		double y, a, t;

		a = (8d*(PI - 3d))/(3d*PI*(4d - PI));

		t = 2d/(PI*a) + Log(1d - Pow(x, 2d))/2d;
		y = m18Sign(x)*Sqrt(Sqrt(Pow(t, 2d) - Log(1d - Pow(x, 2d))/a) - t);

		return y;
	}


	public static double m18FallingFactorial(double x, double n){
		double k, y;

		y = 1d;

		for(k = 0d; k <= n - 1d; k = k + 1d){
			y = y*(x - k);
		}

		return y;
	}


	public static double m18RisingFactorial(double x, double n){
		double k, y;

		y = 1d;

		for(k = 0d; k <= n - 1d; k = k + 1d){
			y = y*(x + k);
		}

		return y;
	}


	public static double m18Hypergeometric(double a, double b, double c, double z, double maxIterations, double precision){
		double y;

		if(Abs(z) >= 0.5){
			y = Pow(1d - z, -a)*m18HypergeometricDirect(a, c - b, c, z/(z - 1d), maxIterations, precision);
		}else{
			y = m18HypergeometricDirect(a, b, c, z, maxIterations, precision);
		}

		return y;
	}


	public static double m18HypergeometricDirect(double a, double b, double c, double z, double maxIterations, double precision){
		double y, yp, n;
		bool done;

		y = 0d;
		done = false;

		for(n = 0d; n < maxIterations && !done; n = n + 1d){
			yp = m18RisingFactorial(a, n)*m18RisingFactorial(b, n)/m18RisingFactorial(c, n)*Pow(z, n)/m18Factorial(n);
			if(Abs(yp) < precision){
				done = true;
			}
			y = y + yp;
		}

		return y;
	}


	public static double m18BernouilliNumber(double n){
		return m18AkiyamaTanigawaAlgorithm(n);
	}


	public static double m18AkiyamaTanigawaAlgorithm(double n){
		double m, j, B;
		double [] A;

		A = new double [(int)(n + 1d)];

		for(m = 0d; m <= n; m = m + 1d){
			A[(int)(m)] = 1d/(m + 1d);
			for(j = m; j >= 1d; j = j - 1d){
				A[(int)(j - 1d)] = j*(A[(int)(j - 1d)] - A[(int)(j)]);
			}
		}

		B = A[0];

		delete(A);

		return B;
	}


	public static void delete(System.Object objectx){
		// C# has garbage collection.
	}
}

}